import hashlib
import os
import pickle
import random
import string
import time

import pandas as pd
import tabulate

import audeer

import audb


random.seed(1)

cache = audeer.mkdir("./cache")


def astype(df, dtype):
    """Convert to desired dataframe dtypes."""
    if dtype == "object":
        # Use `object` to represent strings
        df["archive"] = df["archive"].astype("object")
        df["bit_depth"] = df["bit_depth"].astype("int32")
        df["channels"] = df["channels"].astype("int32")
        df["checksum"] = df["checksum"].astype("object")
        df["duration"] = df["duration"].astype("float64")
        df["format"] = df["format"].astype("object")
        df["removed"] = df["removed"].astype("int32")
        df["sampling_rate"] = df["sampling_rate"].astype("int32")
        df["type"] = df["type"].astype("int32")
        df["version"] = df["version"].astype("object")
        df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
        # Set dtypes in library
        audb.core.define.DEPENDENCY_TABLE = {
            "archive": "object",
            "bit_depth": "int32",
            "channels": "int32",
            "checksum": "object",
            "duration": "float64",
            "format": "object",
            "removed": "int32",
            "sampling_rate": "int32",
            "type": "int32",
            "version": "object",
        }
    elif dtype == "string":
        # Use `string` to represent strings
        df["archive"] = df["archive"].astype("string")
        df["bit_depth"] = df["bit_depth"].astype("int32")
        df["channels"] = df["channels"].astype("int32")
        df["checksum"] = df["checksum"].astype("string")
        df["duration"] = df["duration"].astype("float64")
        df["format"] = df["format"].astype("string")
        df["removed"] = df["removed"].astype("int32")
        df["sampling_rate"] = df["sampling_rate"].astype("int32")
        df["type"] = df["type"].astype("int32")
        df["version"] = df["version"].astype("string")
        df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
        # Set dtypes in library
        audb.core.define.DEPENDENCY_TABLE = {
            "archive": "string",
            "bit_depth": "int32",
            "channels": "int32",
            "checksum": "string",
            "duration": "float64",
            "format": "string",
            "removed": "int32",
            "sampling_rate": "int32",
            "type": "int32",
            "version": "string",
        }
    elif dtype == "pyarrow":
        # Use `pyarrow` to represent all dtypes
        df["archive"] = df["archive"].astype("string[pyarrow]")
        df["bit_depth"] = df["bit_depth"].astype("int32[pyarrow]")
        df["channels"] = df["channels"].astype("int32[pyarrow]")
        df["checksum"] = df["checksum"].astype("string[pyarrow]")
        df["duration"] = df["duration"].astype("float64[pyarrow]")
        df["format"] = df["format"].astype("string[pyarrow]")
        df["removed"] = df["removed"].astype("int32[pyarrow]")
        df["sampling_rate"] = df["sampling_rate"].astype("int32[pyarrow]")
        df["type"] = df["type"].astype("int32[pyarrow]")
        df["version"] = df["version"].astype("string[pyarrow]")
        df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
        # Set dtypes in library
        audb.core.define.DEPENDENCY_TABLE = {
            "archive": "string[pyarrow]",
            "bit_depth": "int32[pyarrow]",
            "channels": "int32[pyarrow]",
            "checksum": "string[pyarrow]",
            "duration": "float64[pyarrow]",
            "format": "string[pyarrow]",
            "removed": "int32[pyarrow]",
            "sampling_rate": "int32[pyarrow]",
            "type": "int32[pyarrow]",
            "version": "string[pyarrow]",
        }
    return df


# === Dependencies pandas.DataFrame ===
data_cache = audeer.path(cache, "df.pkl")
num_rows = 1000000
if not os.path.exists(data_cache):
    bit_depths = [0, 16, 24]
    channels = [0, 1, 2]
    formats = ["csv", "wav", "txt"]
    sampling_rates = [0, 16000, 44100]
    types = [0, 1, 2]
    versions = ["1.0.0", "1.1.0"]
    records = [
        {
            "file": f"file-{n}.wav",
            "archive": f"archive-{n}",
            "bit_depth": random.choices(bit_depths, weights=[0.1, 0.8, 0.1])[0],
            "channels": random.choices(channels, weights=[0.1, 0.8, 0.1])[0],
            "checksum": hashlib.md5(
                pickle.dumps(random.choice(string.ascii_letters))
            ).hexdigest(),
            "duration": 10 * random.random(),
            "format": random.choices(formats, weights=[0.1, 0.8, 0.1])[0],
            "removed": random.choices([0, 1], weights=[0.1, 0.9])[0],
            "sampling_rate": random.choices(sampling_rates, weights=[0.1, 0.8, 0.1])[0],
            "type": random.choices(types, weights=[0.1, 0.8, 0.1])[0],
            "version": random.choices(versions, weights=[0.2, 0.8])[0],
        }
        for n in range(num_rows)
    ]
    df = pd.DataFrame.from_records(records)
    df = df.astype(audb.core.define.DEPENDENDENCY_TABLE)
    df.set_index("file", inplace=True)
    df.index.name = None
    df.index = df.index.astype(audb.core.define.DEPENDENCY_INDEX_DTYPE)
    df.to_pickle(data_cache)


# ===== Benchmark audb.Dependencies =====
deps = audb.Dependencies()
deps.load(data_cache)
file = "file-10.wav"
n_files = 10000
_files = deps._df.index[:n_files].tolist()
dtypes = ["string", "object", "pyarrow"]
results = pd.DataFrame(columns=dtypes)
results.index.name = "method"

for dtype in dtypes:
    deps.load(data_cache)
    deps._df = astype(deps._df, dtype)

    # Check we have the expected dtypes
    # in dependency table
    # and library
    if dtype == "pyarrow":
        expected_dtype = "string[pyarrow]"
    else:
        expected_dtype = dtype
    assert deps._df.archive.dtype == expected_dtype
    assert audb.core.define.DEPENDENCY_TABLE["archive"] == expected_dtype

    method = "Dependencies.__call__()"
    t0 = time.time()
    deps()
    t = time.time() - t0
    results.at[method, dtype] = t

    # Access the index one time.
    # Further calls will be faster
    file in deps

    method = f"Dependencies.__contains__({n_files} files)"
    t0 = time.time()
    [file in deps for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.__get_item__({n_files} files)"
    t0 = time.time()
    [deps[file] for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.__len__()"
    t0 = time.time()
    len(deps)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.__str__()"
    t0 = time.time()
    str(deps)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.archives"
    t0 = time.time()
    deps.archives
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.attachments"
    t0 = time.time()
    deps.attachments
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.attachment_ids"
    t0 = time.time()
    deps.attachment_ids
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.files"
    t0 = time.time()
    deps.files
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.media"
    t0 = time.time()
    deps.media
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.removed_media"
    t0 = time.time()
    deps.removed_media
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.table_ids"
    t0 = time.time()
    deps.table_ids
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies.tables"
    t0 = time.time()
    deps.tables
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.archive({n_files} files)"
    t0 = time.time()
    [deps.archive(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.bit_depth({n_files} files)"
    t0 = time.time()
    [deps.bit_depth(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.channels({n_files} files)"
    t0 = time.time()
    [deps.channels(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.checksum({n_files} files)"
    t0 = time.time()
    [deps.checksum(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.duration({n_files} files)"
    t0 = time.time()
    [deps.duration(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.format({n_files} files)"
    t0 = time.time()
    [deps.format(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.removed({n_files} files)"
    t0 = time.time()
    [deps.removed(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.sampling_rate({n_files} files)"
    t0 = time.time()
    [deps.sampling_rate(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.type({n_files} files)"
    t0 = time.time()
    [deps.type(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies.version({n_files} files)"
    t0 = time.time()
    [deps.version(file) for file in _files]
    t = time.time() - t0
    results.at[method, dtype] = t

    # -------------------------------------------------------------------------
    method = "Dependencies._add_attachment()"
    t0 = time.time()
    deps._add_attachment("attachment.txt", "1.0.0", "archive", "checksum")
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies._add_media({n_files} files)"
    values = [
        (
            f"file-new-{n}.wav",  # file
            f"archive-new-{n}",  # archive
            16,  # bit_depth
            1,  # channels
            f"checksum-{n}",  # checksum
            0.4,  # duration
            "wav",  # format
            0,  # removed
            16000,  # sampling_rate
            1,  # type
            "1.0.0",  # version
        )
        for n in range(n_files)
    ]
    t0 = time.time()
    deps._add_media(values)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._add_meta()"
    t0 = time.time()
    deps._add_meta("db.new-table.csv", "1.0.0", "archive", "checksum")
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._drop()"
    t0 = time.time()
    deps._drop(["file-90000.wav"])
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._remove()"
    t0 = time.time()
    deps._remove(file)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = "Dependencies._update_media()"
    t0 = time.time()
    deps._update_media(values)
    t = time.time() - t0
    results.at[method, dtype] = t

    method = f"Dependencies._update_media_version({n_files} files)"
    t0 = time.time()
    deps._update_media_version([f"file-{n}.wav" for n in range(n_files)], "version")
    t = time.time() - t0
    results.at[method, dtype] = t


# ===== Save results =====
fp_results = audeer.path(cache, "results.csv")
results.to_csv(fp_results)

# ===== Print results =====
table = tabulate.tabulate(results, headers="keys", tablefmt="github", floatfmt=".3f")
print(table)



import os

import audeer
import audformat

import audb


def test_available(repository):
    # Broken database in repo
    name = "non-existing-database"
    path = os.path.join(
        repository.host,
        repository.name,
        name,
    )
    path = audeer.mkdir(path)
    df = audb.available()
    os.rmdir(path)
    assert len(df) == 0

    # Non existing repo
    name = "non-existing-repo"
    audb.config.REPOSITORIES = [
        audb.Repository(name, repository.host, repository.backend)
    ]
    df = audb.available()
    assert len(df) == 0

    # Non existing backend
    audb.config.REPOSITORIES = [
        audb.Repository(repository.name, repository.host, "custom")
    ]
    df = audb.available()
    assert len(df) == 0

    # artifactory backend with non-existing host,
    # and non-support under Python>=3.12
    audb.config.REPOSITORIES = [
        audb.Repository("repo", "https:artifactory.url.com", "artifactory")
    ]
    df = audb.available()
    assert len(df) == 0


def test_available_broken_dataset(private_and_public_repository):
    """Test for listing datasets, including a broken one.

    This uses the public repositories,
    from which ``data-public2``
    includes ``broken-dataset``,
    which has a missing ``db`` folder.

    """
    df = audb.available(only_latest=True)
    assert len(df) > 0
    assert "broken-dataset" not in df


def test_versions(tmpdir, repository):
    """Test versions() for non existing repositories.

    As ``audb.versions()`` does not use ``audb.core.utils._lookup()``
    to get the backend of the corresponding database,
    but goes through all provided backends to collect versions
    of the requested database,
    we need to ensure,
    that it does not crash
    when a repository does not exist.

    See https://github.com/audeering/audb/issues/389.

    """
    # Add non existing repository to the list of configured repositories
    non_existing_repository = audb.Repository(
        name="non-existing-repo",
        host="non-existing-host",
        backend="file-system",
    )
    audb.config.REPOSITORIES += [non_existing_repository]
    # Publish a dataset to an existing repository
    name = "mydb"
    version = "1.0.0"
    build_dir = audeer.mkdir(tmpdir, "build")
    db = audformat.Database(name)
    db.save(build_dir)
    audb.publish(build_dir, version, repository)

    assert audb.versions(name) == [version]



import os

import oyaml as yaml

import audeer

from audb.core.define import CONFIG_FILE
from audb.core.define import DEPRECATED_USER_CONFIG_FILE
from audb.core.define import USER_CONFIG_FILE
from audb.core.repository import Repository


CWD = audeer.script_dir()
global_config_file = os.path.join(CWD, CONFIG_FILE)


def load_configuration_file(config_file: str) -> dict:
    r"""Read configuration from YAML file.

    Args:
        config_file: path to configuration file.
            File doesn't have to exist

    Returns:
        dictionary containing configuration entries

    Raises:
        ValueError: if ``repositories`` section is present,
            but empty
        ValueError: if repository has no ``host``
            ``backend``, or ``name`` key

    """
    if not os.path.exists(config_file):
        return {}

    with open(config_file) as cf:
        config = yaml.load(cf, Loader=yaml.BaseLoader)
        if config is None:
            return {}

    # Check that we have provided a valid repositories configuration
    if "repositories" in config:
        if len(config["repositories"]) == 0:
            raise ValueError(
                "You cannot specify an empty 'repositories:' section "
                f"in the configuration file '{USER_CONFIG_FILE}'."
            )
        for n, repo in enumerate(config["repositories"]):
            if "host" not in repo:
                raise ValueError(
                    f"Your repository is missing a 'host' entry: '{repo}'."
                )
            if "backend" not in repo:
                raise ValueError(
                    f"Your repository is missing a 'backend' entry: '{repo}'."
                )
            if "name" not in repo:
                raise ValueError(
                    f"Your repository is missing a 'name' entry: '{repo}'."
                )

    return config


def load_config() -> dict:
    """Read configuration from configuration files.

    User config values take precedence over global config values
    when the same setting exists in both files.

    """
    # Global config
    config = load_configuration_file(global_config_file)
    # User config
    if os.path.exists(audeer.path(USER_CONFIG_FILE)):
        user_config_file = audeer.path(USER_CONFIG_FILE)
    else:
        user_config_file = audeer.path(DEPRECATED_USER_CONFIG_FILE)
    user_config = load_configuration_file(user_config_file)
    config.update(user_config)
    return config


class config:
    """Get/set configuration values for the :mod:`audb` module.

    The configuration values are read in during module import
    from the :ref:`configuration file <configuration>`
    :file:`~/.config/audb.yaml`.
    You can change the configuration values after import,
    by setting the attributes directly.
    The :ref:`caching <caching>` related configuration values
    can be overwritten by environment variables.

    Examples:
        >>> audb.config.CACHE_ROOT
        '~/audb'
        >>> audb.config.CACHE_ROOT = "~/caches/audb"
        >>> audb.config.CACHE_ROOT
        '~/caches/audb'

    """

    _config = load_config()

    CACHE_ROOT = _config["cache_root"]
    r"""Default user cache folder."""

    REPOSITORIES = [
        Repository(r["name"], r["host"], r["backend"]) for r in _config["repositories"]
    ]
    r"""Repositories, will be iterated in given order.

    A repository is defined by the object :class:`audb.Repository`,
    containing the following attributes:

    * :attr:`audb.Repository.name`: repository name, e.g. ``'data-local'``
    * :attr:`audb.Repository.backend`: backend name, e.g. ``'artifactory'``
    * :attr:`audb.Repository.host`: host name,
      e.g. ``'https://artifactory.audeering.com/artifactory'``

    """

    SHARED_CACHE_ROOT = _config["shared_cache_root"]
    r"""Shared cache folder."""



from audb.core.info import attachments
from audb.core.info import author
from audb.core.info import bit_depths
from audb.core.info import channels
from audb.core.info import description
from audb.core.info import duration
from audb.core.info import files
from audb.core.info import formats
from audb.core.info import header
from audb.core.info import languages
from audb.core.info import license
from audb.core.info import license_url
from audb.core.info import media
from audb.core.info import meta
from audb.core.info import misc_tables
from audb.core.info import organization
from audb.core.info import raters
from audb.core.info import sampling_rates
from audb.core.info import schemes
from audb.core.info import source
from audb.core.info import splits
from audb.core.info import tables
from audb.core.info import usage



import os


# Configuration files
CONFIG_FILE = os.path.join("etc", "audb.yaml")
USER_CONFIG_FILE = "~/.config/audb.yaml"
DEPRECATED_USER_CONFIG_FILE = "~/.audb.yaml"

# Database
DB = "db"
HEADER_FILE = f"{DB}.yaml"

# Dependencies
DEPENDENCY_FILE = f"{DB}.parquet"
r"""Filename and extension of dependency table file."""

CACHED_DEPENDENCY_FILE = f"{DB}.pkl"
r"""Filename and extension of cached dependency table file.

As loading from a pickle file is still faster
than loading from a parquet file,
we are storing the dependency table
as a pickle file in cache.

"""

LEGACY_DEPENDENCY_FILE = f"{DB}.csv"
r"""Filename and extension of legacy dependency table file.

In ``audb`` versions smaller than 1.7.0,
the dependency table was stored in a csv file.

"""

DEPENDENCY_TABLE = {
    # Column name: column dtype
    "archive": "string[pyarrow]",
    "bit_depth": "int32[pyarrow]",
    "channels": "int32[pyarrow]",
    "checksum": "string[pyarrow]",
    "duration": "float64[pyarrow]",
    "format": "string[pyarrow]",
    "removed": "int32[pyarrow]",
    "sampling_rate": "int32[pyarrow]",
    "type": "int32[pyarrow]",
    "version": "string[pyarrow]",
}
r"""Column names and data types of dependency table.

The dependency table is stored in a dataframe
at ``audb.Dependencies._df``,
and contains the specified column names
and data types.

"""

DEPENDENCY_INDEX_DTYPE = "object"
r"""Data type of the dependency table index."""

DEPENDENCY_TYPE = {
    "meta": 0,
    "media": 1,
    "attachment": 2,
}
r"""Types of files stored in a database.

Currently, a database can contain the following files:

* ``"meta"``: tables and misc tables
* ``"media"``: media files, e.g. audio, video, text
* ``"attachment"``: files included as attachments

"""

# Cache lock
CACHED_VERSIONS_TIMEOUT = 10  # Timeout to acquire access to cached versions
LOCK_FILE = ".lock"
TIMEOUT_MSG = "Lock could not be acquired. Timeout exceeded."


# Flavors
class Format:
    r"""Media formats.

    Media formats :meth:`audb.load` can convert to on request.

    """

    WAV = "wav"
    FLAC = "flac"


FORMATS = [Format.WAV, Format.FLAC]
BIT_DEPTHS = [16, 24, 32]
SAMPLING_RATES = [8000, 16000, 22050, 24000, 44100, 48000]

# Progress bar
MAXIMUM_REFRESH_TIME = 1  # force progress bar to update every second



from __future__ import annotations

from collections.abc import Sequence
import os

import audbackend
import audeer
import audformat

from audb.core import define
from audb.core import utils
from audb.core.api import dependencies
from audb.core.api import latest_version
from audb.core.dependencies import Dependencies
from audb.core.load import database_tmp_root
from audb.core.load import load_header_to


def _find_attachments(
    db_root: str,
    deps: Dependencies,
) -> list[str]:
    r"""Find missing attachments."""
    attachments = []

    for file in deps.attachments:
        full_file = os.path.join(db_root, file)
        if not os.path.exists(full_file):
            attachments.append(file)

    return attachments


def _find_media(
    db: audformat.Database,
    db_root: str,
    deps: Dependencies,
    num_workers: int | None,
    verbose: bool,
) -> list[str]:
    r"""Find missing media.

    Collects all media files present in ``db.files``,
    but not in ``db_root``.
    This will find missing files,
    but also altered files
    as those have been deleted
    in a previous step.

    """
    media = []

    def job(file: str):
        if not deps.removed(file):
            full_file = os.path.join(db_root, file)
            if not os.path.exists(full_file):
                media.append(file)

    audeer.run_tasks(
        job,
        params=[([file], {}) for file in db.files],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Find media",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )

    return media


def _find_tables(
    db_header: audformat.Database,
    db_root: str,
    deps: Dependencies,
    num_workers: int | None,
    verbose: bool,
) -> list[str]:
    r"""Find missing tables.

    Collects all tables and misc tables
    present in ``db_header``,
    but not in ``db_root``.
    This will find missing tables,
    but also altered tables
    as those have been deleted
    in a previous step.

    Args:
        db_header: database header
        db_root: database root folder
        deps: database dependencies
        num_workers: number of workers
        verbose: if ``True``, show progress bar

    Returns:
        list of table IDs in ``db_header``,
        not stored in ``db_root``

    """
    tables = []

    def job(table: str):
        if not os.path.exists(
            os.path.join(db_root, f"db.{table}.csv")
        ) and not os.path.exists(os.path.join(db_root, f"db.{table}.parquet")):
            tables.append(table)

    audeer.run_tasks(
        job,
        params=[([table], {}) for table in list(db_header)],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Find tables",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )

    return tables


def _get_attachments(
    paths: Sequence[str],
    db_root: str,
    db_root_tmp: str,
    db_name: str,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    r"""Load attachments from backend."""
    # create folder tree to avoid race condition
    # in os.makedirs when files are unpacked
    utils.mkdir_tree(paths, db_root)
    utils.mkdir_tree(paths, db_root_tmp)

    def job(path: str):
        archive = deps.archive(path)
        version = deps.version(path)
        archive = backend_interface.join("/", db_name, "attachment", archive + ".zip")
        backend_interface.get_archive(
            archive,
            db_root_tmp,
            version,
            tmp_root=db_root_tmp,
        )
        src_path = audeer.path(db_root_tmp, path)
        dst_path = audeer.path(db_root, path)
        audeer.move_file(
            src_path,
            dst_path,
        )

    audeer.run_tasks(
        job,
        params=[([path], {}) for path in paths],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Load attachments",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def _get_media(
    media: list[str],
    db_root: str,
    db_root_tmp: str,
    db_name: str,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    # create folder tree to avoid race condition
    # in os.makedirs when files are unpacked
    utils.mkdir_tree(media, db_root)
    utils.mkdir_tree(media, db_root_tmp)

    # figure out archives
    archives = set()
    for file in media:
        archives.add((deps.archive(file), deps.version(file)))

    def job(archive: str, version: str):
        archive = backend_interface.join("/", db_name, "media", archive + ".zip")
        files = backend_interface.get_archive(
            archive,
            db_root_tmp,
            version,
            tmp_root=db_root_tmp,
        )
        for file in files:
            audeer.move_file(
                os.path.join(db_root_tmp, file),
                os.path.join(db_root, file),
            )

    audeer.run_tasks(
        job,
        params=[([archive, version], {}) for archive, version in archives],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Get media",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def _get_tables(
    tables: list[str],
    db_root: str,
    db_root_tmp: str,
    db_name: str,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    r"""Load table files from backend.

    Args:
        tables: table IDs
        db_root: database root folder
        db_root_tmp: database root temp folder
        db_name: database name
        deps: database dependencies
        backend_interface: backend interface
        num_workers: number of workers
        verbose: if ``True``, show progress bar

    """

    def job(table: str):
        pkl_file = f"db.{table}.pkl"
        csv_file = f"db.{table}.csv"
        parquet_file = f"db.{table}.parquet"

        # If a pickled version of the table exists,
        # we have to remove it to make sure that
        # later on the new CSV tables are loaded.
        # This can happen if we upgrade an existing
        # database to a different version.
        if os.path.exists(os.path.join(db_root, pkl_file)):
            os.remove(os.path.join(db_root, pkl_file))

        # Table file in CSV format is stored in ZIP archive on backend,
        # table file in PARQUET format is stored as PARQUET file on backend.
        if csv_file in deps.tables:
            table_file = csv_file
            remote_file = backend_interface.join("/", db_name, "meta", f"{table}.zip")
            backend_interface.get_archive(
                remote_file,
                db_root_tmp,
                deps.version(table_file),
                tmp_root=db_root_tmp,
            )
        else:
            table_file = parquet_file
            remote_file = backend_interface.join(
                "/", db_name, "meta", f"{table}.parquet"
            )
            backend_interface.get_file(
                remote_file,
                os.path.join(db_root_tmp, table_file),
                deps.version(table_file),
            )

        audeer.move_file(
            os.path.join(db_root_tmp, table_file),
            os.path.join(db_root, table_file),
        )

    audeer.run_tasks(
        job,
        params=[([table], {}) for table in tables],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Get tables",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def _remove_empty_dirs(root):
    r"""Remove directories, fails if it contains non-empty sub-folders."""
    files = os.listdir(root)
    if len(files):
        for file in files:
            full_file = os.path.join(root, file)
            if os.path.isdir(full_file):
                _remove_empty_dirs(full_file)

    os.rmdir(root)


def load_to(
    root: str,
    name: str,
    *,
    version: str = None,
    only_metadata: bool = False,
    pickle_tables: bool = True,
    cache_root: str = None,
    num_workers: int | None = 1,
    verbose: bool = True,
) -> audformat.Database:
    r"""Load database to directory.

    Loads the original state of the database
    to a custom directory.
    No conversion or filtering will be applied.
    If the target folder already contains
    some version of the database,
    it will upgrade to the requested version.
    Unchanged files will be skipped.

    Args:
        root: target directory
        name: name of database
        version: version string, latest if ``None``
        only_metadata: load only header and tables of database
        pickle_tables: if ``True``,
            tables are stored in ``root``
            in their original format
            and as pickle files.
            This allows for faster loading,
            when loading from ``root``
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used.
            Only used to read the dependencies of the requested version
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5
        verbose: show debug messages

    Returns:
        database object

    """
    if version is None:
        version = latest_version(name)

    db_root = audeer.path(root, follow_symlink=True)
    db_root_tmp = database_tmp_root(db_root)

    # remove files with a wrong checksum
    # to ensure we load correct version
    update = os.path.exists(db_root) and os.listdir(db_root)
    audeer.mkdir(db_root)
    deps = dependencies(
        name,
        version=version,
        cache_root=cache_root,
        verbose=verbose,
    )
    if update:
        if only_metadata:
            files = deps.tables
        else:
            files = deps.attachments + deps.files
        for file in files:
            full_file = os.path.join(db_root, file)
            if os.path.exists(full_file):
                checksum = utils.md5(full_file)
                if checksum != deps.checksum(file):
                    if os.path.isdir(full_file):
                        audeer.rmdir(full_file)
                    else:
                        os.remove(full_file)

    # load database header without tables from backend

    db_header, backend_interface = load_header_to(
        db_root_tmp,
        name,
        version,
        overwrite=True,
    )
    db_header.save(db_root_tmp, header_only=True)

    # get altered and new attachments

    if not only_metadata:
        attachments = _find_attachments(
            db_root,
            deps,
        )
        _get_attachments(
            attachments,
            db_root,
            db_root_tmp,
            name,
            deps,
            backend_interface,
            num_workers,
            verbose,
        )

    # get altered and new tables

    tables = _find_tables(db_header, db_root, deps, num_workers, verbose)
    _get_tables(
        tables,
        db_root,
        db_root_tmp,
        name,
        deps,
        backend_interface,
        num_workers,
        verbose,
    )

    # load database

    # move header to root and load database ...
    audeer.move_file(
        os.path.join(db_root_tmp, define.HEADER_FILE),
        os.path.join(db_root, define.HEADER_FILE),
    )
    try:
        db = audformat.Database.load(
            db_root,
            num_workers=num_workers,
            verbose=verbose,
        )
    except (KeyboardInterrupt, Exception):  # pragma: no cover
        # make sure to remove header if user interrupts
        os.remove(os.path.join(db_root, define.HEADER_FILE))
        raise

    # afterwards remove header to avoid the database
    # can be loaded before download is complete
    os.remove(os.path.join(db_root, define.HEADER_FILE))

    # get altered and new media files

    if not only_metadata:
        media = _find_media(db, db_root, deps, num_workers, verbose)
        _get_media(
            media,
            db_root,
            db_root_tmp,
            name,
            deps,
            backend_interface,
            num_workers,
            verbose,
        )

    # save dependencies

    dep_path_tmp = os.path.join(db_root_tmp, define.DEPENDENCY_FILE)
    deps.save(dep_path_tmp)
    audeer.move_file(
        dep_path_tmp,
        os.path.join(db_root, define.DEPENDENCY_FILE),
    )

    # save database and PKL tables

    if pickle_tables:
        # Store database header,
        # and add table as pickle files
        # (tables are already stored in their original format).
        # Uses `num_workers` to save tables in parallel.
        db.save(
            db_root,
            storage_format=audformat.define.TableStorageFormat.PICKLE,
            update_other_formats=False,
            num_workers=num_workers,
            verbose=verbose,
        )
    else:
        # Store database header
        # (tables are already stored)
        db.save(
            db_root,
            header_only=True,
            verbose=verbose,
        )

    # remove the temporal directory
    # to signal all files were correctly loaded
    try:
        _remove_empty_dirs(db_root_tmp)
    except OSError:  # pragma: no cover
        raise RuntimeError(
            "Could not remove temporary directory, "
            "probably there are some leftover files. "
            "This should not happen."
        )

    return db



from audb import info
from audb.core.api import available
from audb.core.api import cached
from audb.core.api import dependencies
from audb.core.api import exists
from audb.core.api import flavor_path
from audb.core.api import latest_version
from audb.core.api import remove_media
from audb.core.api import repository
from audb.core.api import versions
from audb.core.cache import default_cache_root
from audb.core.config import config
from audb.core.dependencies import Dependencies
from audb.core.flavor import Flavor
from audb.core.load import load
from audb.core.load import load_attachment
from audb.core.load import load_media
from audb.core.load import load_table
from audb.core.load_to import load_to
from audb.core.publish import publish
from audb.core.repository import Repository
from audb.core.stream import DatabaseIterator
from audb.core.stream import stream


__all__ = []


# Dynamically get the version of the installed module
try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__name__)
except Exception:  # pragma: no cover
    importlib = None  # pragma: no cover
finally:
    del importlib



import os

import numpy as np
import pandas as pd
import pytest

import audeer
import audformat
import audiofile

import audb


def create_audio_files(root: str, index: pd.Index, *, sampling_rate=16_000):
    r"""Create audio files.

    Given an index with relative paths,
    and a root folder,
    it will create audio files
    for every entry in the index.

    The audio files have a duration of 1 second,
    and have a constant magnitude of 1.

    Args:
        root: root folder of audio files
        index: name of audio files to create in ``root``
        sampling_rate: sampling rate in Hz

    """
    for file in index:
        path = audeer.path(root, file)
        signal = np.ones((1, sampling_rate))
        audiofile.write(path, signal, sampling_rate)


class TestStreaming:
    r"""Test streaming functionality of audb.

    This test tackles ``audb.stream()``,
    and its returned class ``audb.DatabaseIterator``.

    """

    name = "db"
    """Name of test database."""

    version = "1.0.0"
    """Version of test database."""

    seed = 2
    """Seed for random operations."""

    @classmethod
    @pytest.fixture(scope="class", autouse=True, params=["parquet", "csv"])
    def setup(cls, request, tmpdir_factory, persistent_repository):
        r"""Publish a database.

        This creates a database,
        consisting of 5 audio files,
        2 misc tables (1 used as scheme labels),
        and 2 tables (1 fileswise, 1 segmented).

        The database is published to a new repository
        for each ``storage_format``,
        and all tests are run for a given ``storage_format``.

        Args:
            request: request fixture to access params
            tmpdir_factory: tmpdir_factory fixture
            persistent_repository: persistent_repository fixture

        """
        storage_format = request.param

        db_root = tmpdir_factory.mktemp("build")

        db = audformat.Database(cls.name)

        # Misc table for scheme labels
        db.schemes["year-of-birth"] = audformat.Scheme("date")
        db["speaker"] = audformat.MiscTable(
            pd.Index([0, 1, 2, 3, 4], name="speaker", dtype="Int64")
        )
        db["speaker"]["year-of-birth"] = audformat.Column(scheme_id="year-of-birth")
        db["speaker"]["year-of-birth"].set([1910, 1920, 1930, 1940, 1950])
        db.schemes["speaker"] = audformat.Scheme("int", labels="speaker")

        # Misc table
        db.schemes["full-name"] = audformat.Scheme("str")
        db["acronym"] = audformat.MiscTable(
            pd.Index(["CCC"], name="acronym", dtype="string")
        )
        db["acronym"]["full-name"] = audformat.Column(scheme_id="full-name")
        db["acronym"]["full-name"].set(["Concordance Correlation Coefficient"])
        db["acronym"]["speaker"] = audformat.Column(scheme_id="speaker")
        db["acronym"]["speaker"].set([0])

        # Filewise table
        db.schemes["quality"] = audformat.Scheme("int", labels=[1, 2, 3])
        index = audformat.filewise_index([f"file{n}.wav" for n in range(5)])
        create_audio_files(db_root, index)
        db["files"] = audformat.Table(index)
        db["files"]["quality"] = audformat.Column(scheme_id="quality")
        db["files"]["quality"].set([1, 1, 2, 2, 3])
        db["files"]["speaker"] = audformat.Column(scheme_id="speaker")
        db["files"]["speaker"].set([0, 1, 2, 3, 4])

        # Segmented table
        index = audformat.segmented_index(
            files=["file0.wav", "file0.wav", "file1.wav", "file2.wav"],
            starts=[0, 0.5, 0.1, 0.1],
            ends=[0.4, 0.9, 0.8, 0.9],
        )
        db["segments"] = audformat.Table(index)
        db["segments"]["noise"] = audformat.Column()
        db["segments"]["noise"].set(["low", "low", "low", "high"])

        db.save(db_root, storage_format=storage_format)

        # Ensure we start with an empty repository
        # for each storage format
        audeer.rmdir(persistent_repository.host)
        audeer.mkdir(persistent_repository.host, persistent_repository.name)

        audb.publish(db_root, cls.version, persistent_repository)

        yield

        # Clean up persistent repository at end
        audeer.rmdir(persistent_repository.host)
        audeer.mkdir(persistent_repository.host, persistent_repository.name)

    @pytest.mark.parametrize("table", ["files", "segments", "speaker", "acronym"])
    @pytest.mark.parametrize("batch_size", [0, 1, 2, 10])
    def test_batch_size(self, table: str, batch_size: int):
        r"""Test table batching.

        If batch size is 0,
        no batch should be returned.
        If batch size is greater than table length,
        a single batch,
        containing the whole table,
        should be returned.

        Args:
            table: table to stream
            batch_size: number of table rows per batch

        """
        db = audb.stream(
            self.name,
            table,
            version=self.version,
            batch_size=batch_size,
            only_metadata=True,
            full_path=False,
            verbose=False,
        )
        batches = [df for df in db]
        # Create expected dataframes from original table
        expected_df = audb.load_table(
            self.name,
            table,
            version=self.version,
            verbose=False,
        )
        if batch_size == 0:
            assert batches == []
        else:
            pd.testing.assert_frame_equal(pd.concat(batches), expected_df)
        if batch_size > len(expected_df):
            assert len(batches) == 1

    @pytest.mark.parametrize("table", ["files", "acronym"])
    @pytest.mark.parametrize("batch_size", [0, 1, 2, 10])
    @pytest.mark.parametrize("buffer_size", [0, 1, 2, 10])
    def test_buffer_size(self, table: str, batch_size: int, buffer_size: int):
        r"""Test buffer size, when shuffling table batches.

        If batch size is 0 or buffer size is 0,
        no batch should be returned.

        Args:
            table: table to stream
            batch_size: number of table rows per batch
            buffer_size: size of buffer to read table,
                before shuffling

        """
        np.random.seed(self.seed)
        db = audb.stream(
            self.name,
            table,
            version=self.version,
            batch_size=batch_size,
            buffer_size=buffer_size,
            shuffle=True,
            only_metadata=True,
            full_path=False,
            verbose=False,
        )
        batches = [df for df in db]
        # Create expected dataframes from original table
        np.random.seed(self.seed)
        expected_df = audb.load_table(
            self.name,
            table,
            version=self.version,
            verbose=False,
        )
        if batch_size == 0 or buffer_size == 0:
            assert batches == []
        else:
            df = pd.concat(batches)
            # Ensure data is shuffled (besides a buffer size of 1)
            if buffer_size == 1 or len(df) == 1:
                assert list(df.index) == list(expected_df.index)
            else:
                assert list(df.index) != list(expected_df.index)
            # Ensure all index entries appear in shuffled batches
            assert sorted(list(df.index)) == sorted(list(expected_df.index))

    @pytest.mark.parametrize(
        "table, expected_tables, expected_schemes",
        [
            ("speaker", ["speaker"], ["year-of-birth"]),
            (
                "acronym",
                ["acronym", "speaker"],
                ["year-of-birth", "full-name", "speaker"],
            ),
            ("files", ["files", "speaker"], ["year-of-birth", "quality", "speaker"]),
        ],
    )
    def test_db_cleanup(
        self,
        table: str,
        expected_tables: list,
        expected_schemes: list,
    ):
        r"""Test removal of non-selected tables and schemes.

        The database object (``audb.DatabaseIterator``),
        should not contain unneeded tables or schemes.
        If a misc table is used as scheme labels
        in a scheme of the requested table,
        it and its schemes,
        will also be part of the database object.

        Args:
            table: table to stream
            expected_tables: expected tables in database
            expected_schemes: expected schemes in table

        """
        db = audb.stream(
            self.name,
            table,
            version=self.version,
            only_metadata=True,
            verbose=False,
        )
        assert sorted(list(db.tables) + list(db.misc_tables)) == sorted(expected_tables)
        assert list(db.schemes) == sorted(expected_schemes)

    @pytest.mark.parametrize("full_path", [False, True])
    def test_full_path(self, full_path: bool):
        r"""Test full path in tables.

        Args:
            full_path: if ``True``,
                the path to media files
                should start with ``db.root``

        """
        db = audb.stream(
            self.name,
            "files",
            only_metadata=True,
            full_path=full_path,
            verbose=False,
        )
        df = next(db)
        path = df.index[0]
        if full_path:
            assert path == os.path.join(db.root, "file0.wav")
        else:
            assert path == "file0.wav"

    @pytest.mark.parametrize(
        "table, map",
        [
            ("acronym", {"speaker": "year-of-birth"}),
            ("acronym", {"speaker": ["year-of-birth", "speaker"]}),
            ("acronym", {"speaker": ["speaker", "year-of-birth"]}),
            ("files", {"speaker": "year-of-birth"}),
        ],
    )
    def test_map(self, table: str, map: dict):
        r"""Test mapping of scheme labels.

        Args:
            table: table to stream
            map: mapping of column with scheme labels

        """
        db = audb.stream(
            self.name,
            table,
            version=self.version,
            map=map,
            batch_size=16,
            only_metadata=True,
            verbose=False,
        )
        df = next(db)
        expected_df = audb.load_table(
            self.name,
            table,
            version=self.version,
            map=map,
            verbose=False,
        )
        pd.testing.assert_frame_equal(df, expected_df)

    @pytest.mark.parametrize(
        "only_metadata, table, expected_number_of_media_files",
        [
            (True, "files", 5),
            (False, "files", 5),
            (True, "segments", 3),
            (False, "segments", 3),
            (True, "acronym", 0),
            (False, "acronym", 0),
            (True, "speaker", 0),
            (False, "speaker", 0),
        ],
    )
    def test_only_metadata(
        self,
        only_metadata: bool,
        table: str,
        expected_number_of_media_files: int,
    ):
        r"""Test streaming with and without media files.

        Args:
            only_metadata: if ``True``,
                only the table should be streamed
            table: table to stream
            expected_number_of_media_files: expected number of downloaded media files

        """
        db = audb.stream(
            self.name,
            table,
            version=self.version,
            only_metadata=only_metadata,
            verbose=False,
        )
        next(db)
        assert len(db.files) == expected_number_of_media_files
        if not only_metadata:
            for file in db.files:
                assert os.path.exists(file)

    @pytest.mark.parametrize("shuffle", [False, True])
    @pytest.mark.parametrize("table", ["files", "segments", "speaker", "acronym"])
    def test_shuffle(self, shuffle: bool, table: str):
        r"""Test table batch shuffling.

        Args:
            shuffle: if returned table rows should be shuffled
            table: table to stream

        """
        np.random.seed(self.seed)
        db = audb.stream(
            self.name,
            table,
            version=self.version,
            batch_size=16,
            buffer_size=16,
            shuffle=shuffle,
            only_metadata=True,
            full_path=False,
            verbose=False,
        )
        df = next(db)
        # Create expected dataframe from original table
        np.random.seed(self.seed)
        expected_df = audb.load_table(
            self.name,
            table,
            version=self.version,
            verbose=False,
        )
        if shuffle:
            expected_df = expected_df.sample(frac=1)
        pd.testing.assert_frame_equal(df, expected_df)

    @pytest.mark.parametrize(
        "table, error, error_msg",
        [
            (
                "non-existent",
                ValueError,
                "Could not find the table 'non-existent' in db v1.0.0",
            ),
        ],
    )
    def test_errors(self, table: str, error, error_msg: str):
        with pytest.raises(error, match=error_msg):
            audb.stream(
                self.name,
                table,
                version=self.version,
                verbose=False,
            )


def test_database_iterator_error():
    db = audformat.Database("db")
    db["some"] = audformat.Table()
    table = "some"
    error_msg = "Can't instantiate abstract class DatabaseIterator"
    with pytest.raises(TypeError, match=error_msg):
        audb.DatabaseIterator(
            db,
            table,
            version=None,
            map=None,
            batch_size=16,
            shuffle=False,
            buffer_size=100_000,
            only_metadata=False,
            bit_depth=None,
            channels=None,
            format=None,
            mixdown=False,
            sampling_rate=None,
            full_path=False,
            cache_root=None,
            num_workers=1,
            timeout=-1,
            verbose=False,
        )



import os

import audeer

from audb.core.config import config
from audb.core.flavor import Flavor


def database_cache_root(
    name: str,
    version: str,
    cache_root: str = None,
    flavor: Flavor = None,
) -> str:
    r"""Create and return database cache folder.

    Args:
        name: name of database
        version: version of database
        cache_root: path to cache folder
        flavor: flavor of database

    Returns:
        path to cache folder

    """
    if cache_root is None:
        cache_roots = [
            default_cache_root(True),  # check shared cache first
            default_cache_root(False),
        ]
    else:
        cache_roots = [audeer.path(cache_root, follow_symlink=True)]
    for cache_root in cache_roots:
        if flavor is None:
            db_root = os.path.join(cache_root, name, version)
        else:
            db_root = os.path.join(cache_root, flavor.path(name, version))
        if os.path.exists(db_root):
            break

    audeer.mkdir(db_root)
    return db_root


def database_tmp_root(
    db_root: str,
) -> str:
    r"""Create and return temporary database cache folder.

    The temporary cache folder is created under ``db_root + '~'``.

    Args:
        db_root: path to database cache folder

    Returns:
        path to temporary cache folder

    """
    tmp_root = db_root + "~"
    tmp_root = audeer.mkdir(tmp_root)
    return tmp_root


def default_cache_root(
    shared=False,
) -> str:
    r"""Default cache folder.

    If ``shared`` is ``True``,
    returns the path specified
    by the environment variable
    ``AUDB_SHARED_CACHE_ROOT``
    or
    ``audb.config.SHARED_CACHE_ROOT``.
    If ``shared`` is ``False``,
    returns the path specified
    by the environment variable
    ``AUDB_CACHE_ROOT``
    or
    ``audb.config.CACHE_ROOT``.

    Args:
        shared: if ``True`` returns path to shared cache folder

    Returns:
        path normalized by :func:`audeer.path`

    """
    if shared:
        cache = os.environ.get("AUDB_SHARED_CACHE_ROOT") or config.SHARED_CACHE_ROOT
    else:
        cache = os.environ.get("AUDB_CACHE_ROOT") or config.CACHE_ROOT
    return audeer.path(cache, follow_symlink=True)



from collections.abc import Sequence
import os
import warnings

import pyarrow.parquet as parquet

import audbackend
import audeer

from audb.core import define
from audb.core.config import config
from audb.core.repository import Repository


def lookup_backend(
    name: str,
    version: str,
) -> type[audbackend.interface.Base]:
    r"""Return backend of requested database.

    If the database is stored in several repositories,
    only the first one is considered.
    The order of the repositories to look for the database
    is given by :attr:`config.REPOSITORIES`.

    Args:
        name: database name
        version: version string

    Returns:
        backend interface

    Raises:
        RuntimeError: if database is not found

    """
    return _lookup(name, version)[1]


def md5(file: str) -> str:
    r"""MD5 checksum of file.

    PARQUET files are stored non-deterministically.
    To ensure tracking changes to those files correctly,
    the checksum can be provided
    under the key ``b"hash"`` in its metadata,
    e.g. which is done when creating a PARQUET file
    with :meth:`audformat.Table.save`.

    If the key is not present in its metadata,
    or the file is not a PARQUET file
    :func:`audeer.md5` is used to calculate the checksum.

    Args:
        file: file path with extension

    Returns:
        MD5 checksum of file

    """
    ext = audeer.file_extension(file)
    if ext == "parquet":
        metadata = parquet.read_schema(file).metadata
        if b"hash" in metadata:
            return metadata[b"hash"].decode()
    return audeer.md5(file)


def mkdir_tree(
    files: Sequence[str],
    root: str,
):
    r"""Helper function to create folder tree."""
    folders = set()
    for file in files:
        folders.add(os.path.dirname(file))
    for folder in folders:
        audeer.mkdir(root, folder)


def _lookup(
    name: str,
    version: str,
) -> tuple[Repository, type[audbackend.interface.Base]]:
    r"""Helper function to look up database in all repositories.

    Returns repository, version and backend object.

    """
    for repository in config.REPOSITORIES:
        try:
            backend_interface = repository.create_backend_interface()
            backend_interface.backend.open()
        except (audbackend.BackendError, ValueError):
            continue

        header = backend_interface.join("/", name, "db.yaml")
        if backend_interface.exists(header, version, suppress_backend_errors=True):
            return repository, backend_interface
        else:
            backend_interface.backend.close()

    raise RuntimeError(
        f"Cannot find version " f"'{version}' " f"for database " f"'{name}'."
    )


def timeout_warning():
    warnings.warn(
        define.TIMEOUT_MSG,
        category=UserWarning,
    )



import sys

import pytest

import audbackend

import audb


if hasattr(audbackend.backend, "Artifactory"):
    artifactory_backend = audbackend.backend.Artifactory
else:
    artifactory_backend = None


@pytest.mark.parametrize(
    "repository1, repository2, expected",
    [
        (
            audb.Repository("repo", "host", "file-system"),
            audb.Repository("repo", "host", "file-system"),
            True,
        ),
        (
            audb.Repository("repo1", "host", "file-system"),
            audb.Repository("repo2", "host", "file-system"),
            False,
        ),
    ],
)
def test_repository_eq(repository1, repository2, expected):
    """Test the Repository.__eq__().

    Two repository instances are equal,
    if their string representation matches.

    Args:
        repository1: repository instance
        repository2: repository instance
        expected: expected return value of ``Repository.__eq__()``

    """
    assert (repository1 == repository2) == expected


@pytest.mark.parametrize(
    "backend, host, repo, expected",
    [
        (
            "file-system",
            "host",
            "repo",
            "Repository('repo', 'host', 'file-system')",
        ),
    ],
)
def test_repository_repr(backend, host, repo, expected):
    """Test string representation of Repository.

    Args:
        backend: backend name
        host: host
        repo: repository name
        expected: expected string representation

    """
    repository = audb.Repository(repo, host, backend)
    assert str(repository) == expected


@pytest.mark.parametrize(
    "backend, host, repo, expected_backend, expected_interface",
    [
        (
            "file-system",
            "host",
            "repo",
            audbackend.backend.FileSystem,
            audbackend.interface.Versioned,
        ),
        pytest.param(
            "artifactory",
            "host",
            "repo",
            artifactory_backend,
            audbackend.interface.Maven,
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 12),
                reason="No artifactory backend support in Python>=3.12",
            ),
        ),
    ],
)
def test_repository_create_backend_interface(
    backend,
    host,
    repo,
    expected_backend,
    expected_interface,
):
    """Test creation of backend interface.

    Args:
        backend: backend name
        host: host
        repo: repository name
        expected_backend: expected backend class
        expected_interface: expected backend interface class

    """
    repository = audb.Repository(repo, host, backend)
    backend_interface = repository.create_backend_interface()
    assert isinstance(backend_interface, expected_interface)
    assert isinstance(backend_interface.backend, expected_backend)
    assert not backend_interface.backend.opened


@pytest.mark.parametrize(
    "backend, host, repo, expected_error_msg, expected_error",
    [
        (
            "custom",
            "host",
            "repo",
            "'custom' is not a registered backend",
            ValueError,
        ),
        pytest.param(
            "artifactory",
            "host",
            "repo",
            "The 'artifactory' backend is not supported in Python>=3.12",
            ValueError,
            marks=pytest.mark.skipif(
                sys.version_info < (3, 12),
                reason="Should only fail for Python>=3.12",
            ),
        ),
    ],
)
def test_repository_create_backend_interface_errors(
    backend,
    host,
    repo,
    expected_error_msg,
    expected_error,
):
    repository = audb.Repository(repo, host, backend)
    with pytest.raises(expected_error, match=expected_error_msg):
        repository.create_backend_interface()



import os

import numpy as np
import pytest

import audiofile
import audobject

import audb


@pytest.mark.parametrize(
    "bit_depth, channels, format, mixdown, sampling_rate",
    [
        (
            None,
            None,
            None,
            False,
            None,
        ),
        (
            16,
            None,
            audb.core.define.Format.WAV,
            True,
            16000,
        ),
        (
            16,
            None,
            audb.core.define.Format.WAV,
            True,
            24000,
        ),
        pytest.param(
            0,
            None,
            audb.core.define.Format.WAV,
            True,
            16000,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            16,
            None,
            "bad",
            True,
            16000,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        pytest.param(
            16,
            None,
            audb.core.define.Format.WAV,
            True,
            0,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_init(bit_depth, channels, format, mixdown, sampling_rate):
    flavor = audb.Flavor(
        bit_depth=bit_depth,
        channels=channels,
        format=format,
        mixdown=mixdown,
        sampling_rate=sampling_rate,
    )
    flavor_s = flavor.to_yaml_s()
    flavor_2 = audobject.from_yaml_s(flavor_s)
    assert isinstance(flavor_2, audb.Flavor)
    assert flavor.id == flavor_2.id


def test_deprecated_sampling_rate():
    """Test a sampling rate of 22500 Hz raises user warning."""
    message = "removed with version 1.13.0"
    with pytest.warns(UserWarning, match=message):
        audb.Flavor(sampling_rate=22500)


@pytest.mark.parametrize(
    "format",
    [
        audb.core.define.Format.WAV,
        audb.core.define.Format.FLAC,
    ],
)
def test_destination(format):
    flavor = audb.Flavor(format=format)
    filename = "wav/audio1.wav"
    expected_filename = f"{filename[:-4]}.{format}"
    assert flavor.destination(filename) == expected_filename


@pytest.mark.parametrize(
    "bit_depth_in, channels_in, format_in, sampling_rate_in, flavor, "
    "bit_depth_out, channels_out, format_out, sampling_rate_out",
    [
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16.0,
            1,
            audb.core.define.Format.WAV,
            16000.0,
            audb.Flavor(),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(bit_depth=32),
            32,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(format=audb.core.define.Format.FLAC),
            16,
            1,
            audb.core.define.Format.FLAC,
            16000,
        ),
        pytest.param(
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(format=audb.core.define.Format.FLAC),
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(mixdown=True),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            3,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(mixdown=True),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(channels=0),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(channels=1),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(channels=[0, 1]),
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            3,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(channels=[0, -1]),
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(channels=[0, 2]),
            16,
            2,
            audb.core.define.Format.WAV,
            16000,
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(sampling_rate=8000),
            16,
            1,
            audb.core.define.Format.WAV,
            8000,
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(
                bit_depth=24,
                format=audb.core.define.Format.FLAC,
                sampling_rate=8000,
            ),
            24,
            1,
            audb.core.define.Format.FLAC,
            8000,
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(
                bit_depth=24,
                format=audb.core.define.Format.FLAC,
                sampling_rate=8000,
            ),
            24,
            1,
            audb.core.define.Format.FLAC,
            8000,
        ),
        (
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
            audb.Flavor(
                bit_depth=24,
                format=audb.core.define.Format.FLAC,
                sampling_rate=8000,
            ),
            24,
            1,
            audb.core.define.Format.FLAC,
            8000,
        ),
        (
            16,
            1,
            "mp3",
            16000,
            audb.Flavor(format=audb.core.define.Format.WAV),
            16,
            1,
            audb.core.define.Format.WAV,
            16000,
        ),
        # Cannot convert MP3 files
        pytest.param(
            16,
            1,
            "mp3",
            16000,
            audb.Flavor(bit_depth=24),
            24,
            1,
            "mp3",
            16000,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(
            16,
            1,
            "mp3",
            16000,
            audb.Flavor(channels=[0, 0]),
            16,
            2,
            "mp3",
            16000,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
        pytest.param(
            16,
            1,
            "mp3",
            16000,
            audb.Flavor(sampling_rate=8000),
            16,
            1,
            "mp3",
            8000,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
    ],
)
def test_call(
    tmpdir,
    bit_depth_in,
    channels_in,
    format_in,
    sampling_rate_in,
    flavor,
    bit_depth_out,
    channels_out,
    format_out,
    sampling_rate_out,
):
    file_in = os.path.join(tmpdir, "in." + format_in)
    file_out = os.path.join(tmpdir, "out." + format_out)

    signal = np.zeros((channels_in, int(sampling_rate_in)), np.float32)
    if format_in == "mp3":
        audiofile.write(
            f"{file_in[:-4]}.wav",
            signal,
            sampling_rate_in,
            int(bit_depth_in),
        )
        os.rename(f"{file_in[:-4]}.wav", file_in)
    else:
        audiofile.write(
            file_in,
            signal,
            int(sampling_rate_in),
            int(bit_depth_in),
        )

    flavor(file_in, file_out)
    assert audiofile.bit_depth(file_out) == bit_depth_out
    assert audiofile.channels(file_out) == channels_out
    assert audiofile.sampling_rate(file_out) == sampling_rate_out



import os

import numpy as np
import pandas as pd
import pytest

import audeer
import audformat.testing
import audiofile

import audb


DB_NAME = "test_info"
DB_VERSION = "1.0.0"


@pytest.fixture(
    scope="module",
    autouse=True,
)
def db(tmpdir_factory, persistent_repository):
    r"""Publish a single database.

    Returns:
        database object

    """
    # creat db

    db = audformat.Database(
        DB_NAME,
        source="https://audeering.github.io/audb/",
        usage=audformat.define.Usage.UNRESTRICTED,
        languages=["de", "English"],
        description="audb.info unit test database",
        meta={"foo": "bar"},
    )
    db.media["media"] = audformat.Media()
    db.schemes["scheme1"] = audformat.Scheme()
    db.splits["split"] = audformat.Split()
    db.raters["rater"] = audformat.Rater()
    db.attachments["attachment"] = audformat.Attachment("file.txt")
    db["table1"] = audformat.Table(
        audformat.filewise_index(
            ["f11.wav", "f12.wav", "f13.wav"],
        ),
        media_id="media",
        split_id="split",
    )
    db["table1"]["column"] = audformat.Column(
        scheme_id="scheme1",
        rater_id="rater",
    )
    db["table2"] = audformat.Table(
        audformat.segmented_index(
            ["f21.wav", "f22.wav", "f22.wav"],
            [0, 0, 0.5],
            [1, 0.5, 1],
        ),
        media_id="media",
        split_id="split",
    )
    db["table2"]["column"] = audformat.Column(
        scheme_id="scheme1",
        rater_id="rater",
    )
    db["misc-in-scheme"] = audformat.MiscTable(pd.Index([0, 1], name="idx"))
    db["misc-not-in-scheme"] = audformat.MiscTable(pd.Index([0, 1], name="idx"))
    db.schemes["scheme2"] = audformat.Scheme(
        "int",
        labels="misc-in-scheme",
    )

    # create db + audio files

    db_root = tmpdir_factory.mktemp(DB_VERSION)
    sampling_rate = 8000
    audeer.touch(db_root, db.attachments["attachment"].path)
    for table in list(db.tables):
        for file in db[table].files:
            audiofile.write(
                os.path.join(db_root, file),
                np.zeros((1, sampling_rate)),
                sampling_rate,
            )

    db.save(db_root)

    # publish db

    audb.publish(
        db_root,
        DB_VERSION,
        persistent_repository,
        verbose=False,
    )

    return db


def test_attachemnts(db):
    assert str(audb.info.attachments(DB_NAME)) == str(db.attachments)


def test_author(db):
    assert audb.info.author(DB_NAME) == db.author


def test_header(db):
    # Load header without loading misc tables
    header = audb.info.header(DB_NAME, load_tables=False)
    assert str(header) == str(db)
    error_msg = "No file found for table with path"
    with pytest.raises(RuntimeError, match=error_msg):
        assert 0 in header.schemes["scheme2"]
    # Load header with tables
    header = audb.info.header(DB_NAME)
    assert str(header) == str(db)
    assert 0 in header.schemes["scheme2"]


def test_bit_depths():
    deps = audb.dependencies(DB_NAME, version=DB_VERSION)
    assert audb.info.bit_depths(DB_NAME) == {
        deps.bit_depth(file) for file in deps.media if deps.bit_depth(file)
    }


def test_channels():
    deps = audb.dependencies(DB_NAME, version=DB_VERSION)
    assert audb.info.channels(DB_NAME) == {
        deps.channels(file) for file in deps.media if deps.channels(file)
    }


def test_description(db):
    assert audb.info.description(DB_NAME) == db.description


@pytest.mark.parametrize(
    "tables, media",
    [
        (None, None),
        ([], None),
        (None, []),
        ("", ""),
        ("table1", None),
        ("misc-in-scheme", None),
        ("misc-not-in-scheme", None),
        (None, ["f11.wav", "f12.wav"]),
        ("table1", ["f11.wav", "f12.wav"]),
        # Error as tables and media do not overlap
        pytest.param(
            "table2",
            ["f11.wav", "f12.wav"],
            marks=pytest.mark.xfail(raises=ValueError),
        ),
    ],
)
def test_duration(tables, media):
    deps = audb.dependencies(DB_NAME, version=DB_VERSION)
    duration = audb.info.duration(DB_NAME, tables=tables, media=media)
    db = audb.load(
        DB_NAME,
        version=DB_VERSION,
        tables=tables,
        media=media,
        full_path=False,
        verbose=False,
    )
    expected_duration = pd.to_timedelta(
        sum([deps.duration(file) for file in db.files]),
        unit="s",
    )
    assert duration == expected_duration


def test_formats():
    deps = audb.dependencies(DB_NAME, version=DB_VERSION)
    assert audb.info.formats(DB_NAME) == {deps.format(file) for file in deps.media}


def test_languages(db):
    assert audb.info.languages(DB_NAME) == db.languages


def test_license(db):
    assert audb.info.license(DB_NAME) == db.license


def test_license_url(db):
    assert audb.info.license_url(DB_NAME) == db.license_url


def test_media(db):
    assert str(audb.info.media(DB_NAME)) == str(db.media)


def test_meta(db):
    assert audb.info.meta(DB_NAME) == db.meta


def test_misc_tables(db):
    assert str(audb.info.misc_tables(DB_NAME)) == str(db.misc_tables)


def test_organization(db):
    assert audb.info.organization(DB_NAME) == db.organization


def test_raters(db):
    assert str(audb.info.raters(DB_NAME)) == str(db.raters)


def test_sampling_rates():
    deps = audb.dependencies(DB_NAME, version=DB_VERSION)
    assert audb.info.sampling_rates(DB_NAME) == {
        deps.sampling_rate(file) for file in deps.media if deps.sampling_rate(file)
    }


def test_schemes(db):
    # Load schemes without loading misc tables
    schemes = audb.info.schemes(DB_NAME, load_tables=False)
    assert str(schemes) == str(db.schemes)
    error_msg = "No file found for table with path"
    with pytest.raises(RuntimeError, match=error_msg):
        assert 0 in schemes["scheme2"]
    # Load header with tables
    schemes = audb.info.schemes(DB_NAME)
    str(schemes) == str(db.schemes)
    assert 0 in schemes["scheme2"]


def test_splits(db):
    assert str(audb.info.splits(DB_NAME)) == str(db.splits)


def test_source(db):
    assert audb.info.source(DB_NAME) == db.source


def test_tables(db):
    assert str(audb.info.tables(DB_NAME)) == str(db.tables)


def test_usage(db):
    assert audb.info.usage(DB_NAME) == db.usage



from collections.abc import Sequence

import pandas as pd

import audformat

from audb.core import define
from audb.core.api import dependencies
from audb.core.load import filtered_dependencies
from audb.core.load import load_header
from audb.core.load import load_table


def attachments(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict[str, audformat.Attachment]:
    """Attachment(s) of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        attachments of database

    Examples:
        >>> list(audb.info.attachments("emodb", version="1.4.1"))
        ['bibtex']

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.attachments


def author(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """Author(s) of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        author(s) of database

    Examples:
        >>> audb.info.author("emodb", version="1.4.1")
        'Felix Burkhardt, Astrid Paeschke, Miriam Rolfes, Walter Sendlmeier, Benjamin Weiss'

    """  # noqa: E501
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.author


def bit_depths(
    name: str,
    *,
    version: str = None,
    tables: Sequence = None,
    media: Sequence = None,
    cache_root: str = None,
) -> set[int]:
    """Media bit depth.

    Args:
        name: name of database
        version: version of database
        tables: include only tables matching the regular expression or
            provided in the list
        media: include only media matching the regular expression or
            provided in the list
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        bit depths

    Raises:
        ValueError: if table or media is requested
            that is not part of the database

    Examples:
        >>> audb.info.bit_depths("emodb", version="1.4.1")
        {16}

    """
    df = filtered_dependencies(name, version, media, tables, cache_root)
    return set(df[df.type == define.DEPENDENCY_TYPE["media"]].bit_depth)


def channels(
    name: str,
    *,
    version: str = None,
    tables: Sequence = None,
    media: Sequence = None,
    cache_root: str = None,
) -> set[int]:
    """Media channels.

    Args:
        name: name of database
        version: version of database
        tables: include only tables matching the regular expression or
            provided in the list
        media: include only media matching the regular expression or
            provided in the list
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        channel numbers

    Raises:
        ValueError: if table or media is requested
            that is not part of the database

    Examples:
        >>> audb.info.channels("emodb", version="1.4.1")
        {1}

    """
    df = filtered_dependencies(name, version, media, tables, cache_root)
    return set(df[df.type == define.DEPENDENCY_TYPE["media"]].channels)


def description(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """Description of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        description of database

    Examples:
        >>> desc = audb.info.description("emodb", version="1.4.1")
        >>> desc.split(".")[0]  # show first sentence
        'Berlin Database of Emotional Speech'

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.description


def duration(
    name: str,
    *,
    version: str = None,
    tables: Sequence = None,
    media: Sequence = None,
    cache_root: str = None,
) -> pd.Timedelta:
    """Total media duration.

    Args:
        name: name of database
        version: version of database
        tables: include only tables matching the regular expression or
            provided in the list
        media: include only media matching the regular expression or
            provided in the list
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        duration

    Raises:
        ValueError: if table or media is requested
            that is not part of the database

    Examples:
        >>> audb.info.duration("emodb", version="1.4.1")
        Timedelta('0 days 00:24:47.092187500')
        >>> audb.info.duration("emodb", version="1.4.1", media=["wav/03a01Fa.wav"])
        Timedelta('0 days 00:00:01.898250')

    """
    df = filtered_dependencies(name, version, media, tables, cache_root)
    return pd.to_timedelta(
        df[df.type == define.DEPENDENCY_TYPE["media"]].duration.sum(),
        unit="s",
    )


def files(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> list[str]:
    """Media files included in the database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        media files

    Examples:
        >>> audb.info.files("emodb", version="1.4.1")[:2]
        ['wav/03a01Fa.wav', 'wav/03a01Nc.wav']

    """
    deps = dependencies(name, version=version, cache_root=cache_root)
    return deps.media


def formats(
    name: str,
    *,
    version: str = None,
    tables: Sequence = None,
    media: Sequence = None,
    cache_root: str = None,
) -> set[str]:
    """Media formats.

    Args:
        name: name of database
        version: version of database
        tables: include only tables matching the regular expression or
            provided in the list
        media: include only media matching the regular expression or
            provided in the list
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        formats

    Raises:
        ValueError: if table or media is requested
            that is not part of the database

    Examples:
        >>> audb.info.formats("emodb", version="1.4.1")
        {'wav'}

    """
    df = filtered_dependencies(name, version, media, tables, cache_root)
    return set(df[df.type == define.DEPENDENCY_TYPE["media"]].format)


def header(
    name: str,
    *,
    version: str = None,
    load_tables: bool = True,
    cache_root: str = None,
) -> audformat.Database:
    r"""Load header of database.

    Args:
        name: name of database
        version: version of database
        load_tables: if ``True``
            downloads misc tables
            used by schemes
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        database object without table data

    Examples:
        >>> db = audb.info.header("emodb", version="1.4.1")
        >>> db.name
        'emodb'

    """
    db = load_header(name, version=version, cache_root=cache_root)
    if load_tables:
        for scheme in db.schemes.values():
            if scheme.uses_table:
                table = scheme.labels
                load_table(
                    name,
                    table,
                    version=version,
                    cache_root=cache_root,
                    verbose=False,
                )
    return db


def languages(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> list[str]:
    """Languages of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        languages of database

    Examples:
        >>> audb.info.languages("emodb", version="1.4.1")
        ['deu']

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.languages


def license(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """License of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        license of database

    Examples:
        >>> audb.info.license("emodb", version="1.4.1")
        'CC0-1.0'

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.license


def license_url(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """License URL of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        license URL of database

    Examples:
        >>> audb.info.license_url("emodb", version="1.4.1")
        'https://creativecommons.org/publicdomain/zero/1.0/'

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.license_url


def media(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict[str, audformat.Media]:
    """Audio and video media of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        media of database

    Examples:
        >>> audb.info.media("emodb", version="1.4.1")
        microphone:
            {type: other, format: wav, channels: 1, sampling_rate: 16000}

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.media


def meta(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict:
    """Meta information of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        meta information of database

    Examples:
        >>> audb.info.meta("emodb", version="1.4.1")
        pdf:
          http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.130.8506&rep=rep1&type=pdf

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.meta


def misc_tables(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict[str, audformat.MiscTable]:
    """Miscellaneous tables of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        miscellaneous tables of database

    Examples:
        >>> list(audb.info.misc_tables("emodb", version="1.4.1"))
        ['speaker']

    """  # noqa: E501
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.misc_tables


def organization(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """Organization responsible for database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        organization responsible for database

    Examples:
        >>> audb.info.organization("emodb", version="1.4.1")
        'audEERING'

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.organization


def raters(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict[str, audformat.Rater]:
    """Raters contributed to database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        raters of database

    Examples:
        >>> audb.info.raters("emodb", version="1.4.1")
        gold:
            {type: human}

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.raters


def sampling_rates(
    name: str,
    *,
    version: str = None,
    tables: Sequence = None,
    media: Sequence = None,
    cache_root: str = None,
) -> set[int]:
    """Media sampling rates.

    Args:
        name: name of database
        version: version of database
        tables: include only tables matching the regular expression or
            provided in the list
        media: include only media matching the regular expression or
            provided in the list
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        sampling rates

    Raises:
        ValueError: if table or media is requested
            that is not part of the database

    Examples:
        >>> audb.info.sampling_rates("emodb", version="1.4.1")
        {16000}

    """
    df = filtered_dependencies(name, version, media, tables, cache_root)
    return set(df[df.type == define.DEPENDENCY_TYPE["media"]].sampling_rate)


def schemes(
    name: str,
    *,
    version: str = None,
    load_tables: bool = True,
    cache_root: str = None,
) -> dict[str, audformat.Scheme]:
    """Schemes of database.

    Args:
        name: name of database
        version: version of database
        load_tables: if ``True``
            downloads misc tables
            used by schemes
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        schemes of database

    Examples:
        >>> list(audb.info.schemes("emodb", version="1.4.1"))
        ['age', 'confidence', 'duration', 'emotion', 'gender', 'language', 'speaker', 'transcription']

    """  # noqa: E501
    db = header(
        name,
        version=version,
        load_tables=load_tables,
        cache_root=cache_root,
    )
    return db.schemes


def source(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """Source of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        source of database

    Examples:
        >>> audb.info.source("emodb", version="1.4.1")
        'http://emodb.bilderbar.info/download/download.zip'

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.source


def splits(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict[str, audformat.Split]:
    """Splits of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        splits of database

    Examples:
        >>> audb.info.splits("emodb", version="1.4.1")
        test:
          {description: Unofficial speaker-independent test split, type: test}
        train:
          {description: Unofficial speaker-independent train split, type: train}

    """  # noqa: E501
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.splits


def tables(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> dict[str, audformat.Table]:
    """Tables of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        tables of database

    Examples:
        >>> list(audb.info.tables("emodb", version="1.4.1"))
        ['emotion', 'emotion.categories.test.gold_standard', 'emotion.categories.train.gold_standard', 'files']

    """  # noqa: E501
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.tables


def usage(
    name: str,
    *,
    version: str = None,
    cache_root: str = None,
) -> str:
    """Usage of database.

    Args:
        name: name of database
        version: version of database
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used

    Returns:
        usage of database

    Examples:
        >>> audb.info.usage("emodb", version="1.4.1")
        'unrestricted'

    """
    db = header(
        name,
        version=version,
        load_tables=False,
        cache_root=cache_root,
    )
    return db.usage



import threading
import time

import filelock

import audeer

from audb.core.lock import FolderLock


event = threading.Event()


def job(lock, wait, sleep):
    if wait:
        event.wait()  # wait for another thread to enter the lock
    try:
        with lock:
            if not wait:
                event.set()  # notify waiting threads to enter the lock
            time.sleep(sleep)
    except filelock.Timeout:
        return 0
    return 1


def test_lock(tmpdir):
    # create two lock folders

    lock_folders = [audeer.mkdir(tmpdir, str(idx)) for idx in range(2)]

    # lock 1 and 2

    lock_1 = FolderLock(lock_folders[0])
    lock_2 = FolderLock(lock_folders[1])

    event.clear()
    result = audeer.run_tasks(
        job,
        [
            ([lock_1, False, 0], {}),
            ([lock_2, False, 0], {}),
        ],
        num_workers=2,
    )
    assert result == [1, 1]

    # lock 1, 2 and 1+2

    lock_1 = FolderLock(lock_folders[0])
    lock_2 = FolderLock(lock_folders[1])
    lock_12 = FolderLock(lock_folders)

    result = audeer.run_tasks(
        job,
        [
            ([lock_1, False, 0], {}),
            ([lock_2, False, 0], {}),
            ([lock_12, False, 0], {}),
        ],
        num_workers=3,
    )
    assert result == [1, 1, 1]

    # lock 1, then 1+2 + wait

    lock_1 = FolderLock(lock_folders[0])
    lock_12 = FolderLock(lock_folders)

    event.clear()
    result = audeer.run_tasks(
        job,
        [
            ([lock_1, False, 0.2], {}),
            ([lock_12, True, 0], {}),
        ],
        num_workers=2,
    )
    assert result == [1, 1]

    # lock 1, then 1+2 + timeout

    lock_1 = FolderLock(lock_folders[0])
    lock_12 = FolderLock(lock_folders, timeout=0)

    event.clear()
    result = audeer.run_tasks(
        job,
        [
            ([lock_1, False, 0.2], {}),
            ([lock_12, True, 0], {}),
        ],
        num_workers=2,
    )
    assert result == [1, 0]

    # lock 1+2, then 1 + wait

    lock_1 = FolderLock(lock_folders[0])
    lock_12 = FolderLock(lock_folders)

    event.clear()
    result = audeer.run_tasks(
        job,
        [
            ([lock_1, True, 0], {}),
            ([lock_12, False, 0.2], {}),
        ],
        num_workers=2,
    )
    assert result == [1, 1]

    # lock 1+2, then 1 + timeout

    lock_1 = FolderLock(lock_folders[0], timeout=0)
    lock_12 = FolderLock(lock_folders)

    event.clear()
    result = audeer.run_tasks(
        job,
        [
            ([lock_1, True, 0], {}),
            ([lock_12, False, 0.2], {}),
        ],
        num_workers=2,
    )
    assert result == [0, 1]

    # lock 1+2, then 1 + wait and 2 + timeout

    lock_1 = FolderLock(lock_folders[0])
    lock_2 = FolderLock(lock_folders[1], timeout=0)
    lock_12 = FolderLock(lock_folders)

    event.clear()
    result = audeer.run_tasks(
        job,
        [
            ([lock_1, True, 0], {}),
            ([lock_2, True, 0], {}),
            ([lock_12, 0, 0.2], {}),
        ],
        num_workers=3,
    )
    assert result == [1, 0, 1]



from __future__ import annotations

import collections
from collections.abc import Mapping
import os
import re
import shutil
import tempfile

import audbackend
import audeer
import audformat
import audiofile

from audb.core import define
from audb.core import utils
from audb.core.api import dependencies
from audb.core.dependencies import Dependencies
from audb.core.dependencies import upload_dependencies
from audb.core.repository import Repository


def _check_for_duplicates(
    db: audformat.Database,
    num_workers: int,
    verbose: bool,
):
    r"""Ensures tables do not contain duplicated index entries."""

    def job(table_id):
        audformat.assert_no_duplicates(db[table_id]._df)

    table_ids = list(db)
    audeer.run_tasks(
        job,
        params=[([table_id], {}) for table_id in table_ids],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Check tables for duplicates",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def _check_for_missing_media(
    db: audformat.Database,
    db_root: str,
    db_root_files: set[str],
    deps: Dependencies,
):
    r"""Check for media that is not in root and not in dependencies."""
    db_files = db.files
    deps_files = deps.media

    db_files_not_in_deps = set(db_files) - set(deps_files)
    missing_files = db_files_not_in_deps - db_root_files

    if len(missing_files) > 0:
        missing_files = sorted(list(missing_files))
        number_of_presented_files = 20
        error_msg = (
            f"The following "
            f"{len(missing_files)} "
            f"files are referenced in tables "
            f"that cannot be found on disk "
            f"and are not yet part of the database: "
            f"{missing_files[:number_of_presented_files]}"
        )
        if len(missing_files) <= number_of_presented_files:
            error_msg += "."
        else:
            error_msg = error_msg[:-1]
            error_msg += ", ...]."
        raise RuntimeError(error_msg)


def _find_attachments(
    db: audformat.Database,
    db_root: str,
    version: str,
    deps: Dependencies,
    verbose: bool,
) -> list[str]:
    r"""Find altered, new or removed attachments and update 'deps'."""
    # drop removed attachments from dependency table
    removed_attachments = [
        deps._df.index[deps._df.archive == attachment_id][0]
        for attachment_id in deps.attachment_ids
        if attachment_id not in db.attachments
    ]
    deps._drop(removed_attachments)

    # check attachments are valid
    db_files = list(db.files) + [f"db.{t}.csv" for t in db.tables]
    for attachment_id in db.attachments:
        path = db.attachments[attachment_id].path
        for file in db_files:
            if file.startswith(path):
                raise RuntimeError(
                    "An attachment must not "
                    "overlap with media or tables. "
                    f"But attachment '{attachment_id}' "
                    f"contains '{file}'."
                )

        path = audeer.path(db.root, db.attachments[attachment_id].path)
        if os.path.exists(path):
            if os.path.isdir(path):
                folders = audeer.list_dir_names(
                    path,
                    recursive=True,
                    hidden=True,
                    basenames=True,
                )
                for folder in folders:
                    if not os.listdir(audeer.path(path, folder)):
                        raise RuntimeError(
                            "An attachment must not "
                            "contain empty sub-folders. "
                            f"But attachment '{attachment_id}' "
                            f"contains the empty sub-folder '{folder}'."
                        )

            files = db.attachments[attachment_id].files
            if len(files) == 0:
                raise RuntimeError(
                    "An attached folder must "
                    "contain at least one file. "
                    f"But attachment '{attachment_id}' "
                    "points to an empty folder."
                )

    # add dependencies to new or updated attachments
    attachment_ids = []
    for attachment_id in audeer.progress_bar(
        list(db.attachments),
        desc="Find attachments",
        disable=not verbose,
    ):
        # use one archive per attachment ID
        path = db.attachments[attachment_id].path
        if not os.path.exists(audeer.path(db_root, path)):
            if path not in deps:
                # Raise FileNotFoundError
                #
                # Attachment is not in deps,
                # but its path does not exist on disk either.
                # We call its `files` property
                # which raises a FileNotFoundError in this case
                db.attachments[attachment_id].files
        else:
            checksum = utils.md5(audeer.path(db_root, path))
            if path not in deps or checksum != deps.checksum(path):
                deps._add_attachment(
                    file=path,
                    version=version,
                    archive=attachment_id,
                    checksum=checksum,
                )
                attachment_ids.append(attachment_id)

    return list(attachment_ids)


def _find_media(
    db: audformat.Database,
    db_root: str,
    db_root_files: set[str],
    version: str,
    deps: Dependencies,
    archives: Mapping[str, str],
    num_workers: int,
    verbose: bool,
) -> set[str]:
    r"""Find archives with new, altered or removed media and update 'deps'."""
    media_archives = set()
    db_media = set(db.files)

    # release dependencies to removed media
    # and select according archives for upload
    removed_files = set(deps.media) - db_media
    for file in removed_files:
        media_archives.add(deps.archive(file))
    deps._drop(removed_files)

    # limit to relevant media
    db_media_in_root = db_media.intersection(db_root_files)

    # update version of altered media and insert new ones

    def job(file):
        path = os.path.join(db_root, file)
        if file not in deps:
            # assert lowercase file extensions
            ext = audeer.file_extension(file)
            if ext.lower() != ext:
                raise RuntimeError(
                    "The file extension of a media file must be lowercase, "
                    f"but '{file}' includes at least one uppercase letter."
                )
            checksum = audeer.md5(path)
            if file in archives:
                archive = archives[file]
            else:
                archive = audeer.uid(from_string=file.replace("\\", "/"))
            values = _media_values(
                db_root,
                file,
                version,
                archive,
                checksum,
            )
            add_media.append(values)
        elif not deps.removed(file):
            checksum = audeer.md5(path)
            if checksum != deps.checksum(file):
                archive = deps.archive(file)
                values = _media_values(
                    db_root,
                    file,
                    version,
                    archive,
                    checksum,
                )
                update_media.append(values)

    add_media = []
    update_media = []
    audeer.run_tasks(
        job,
        params=[([file], {}) for file in db_media_in_root],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Find media",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )
    # Add updated and new media to dependencies
    # and sort them by paths
    # as they can be returned in random order
    # by the discovery job.
    # Sorting enforces determinism in the
    # generation of the dependencies table
    if update_media:
        update_media = sorted(update_media, key=lambda x: x[0])
        deps._update_media(update_media)
    if add_media:
        add_media = sorted(add_media, key=lambda x: x[0])
        deps._add_media(add_media)

    # select archives with new or altered files for upload
    for file in deps.media:
        if not deps.removed(file) and deps.version(file) == version:
            media_archives.add(deps.archive(file))

    return media_archives


def _find_tables(
    db: audformat.Database,
    db_root: str,
    version: str,
    deps: Dependencies,
    verbose: bool,
) -> list[str]:
    r"""Find altered, new or removed tables and update 'deps'."""
    table_ids = list(db)
    # PARQUET is default table,
    # CSV file is ignored
    # if it exists as well
    table_files = [
        f"db.{table}.parquet"
        if os.path.exists(os.path.join(db_root, f"db.{table}.parquet"))
        else f"db.{table}.csv"
        for table in table_ids
    ]

    # release dependencies to removed tables
    deps._drop(set(deps.tables) - set(table_files))

    # new database tables
    tables = []
    for table, file in audeer.progress_bar(
        zip(table_ids, table_files),
        desc="Find tables",
        disable=not verbose,
    ):
        checksum = utils.md5(os.path.join(db_root, file))
        if file not in deps or checksum != deps.checksum(file):
            deps._add_meta(file, version, checksum)
            tables.append(table)

    return tables


def _get_root_files(
    db_root: str,
) -> set[str]:
    r"""Return list of files in root directory."""
    db_root_files = audeer.list_file_names(
        db_root,
        basenames=True,
        recursive=True,
    )
    if os.name == "nt":  # pragma: no cover
        # convert '\\' to '/'
        db_root_files = [file.replace("\\", "/") for file in db_root_files]

    return set(db_root_files)


def _media_values(
    root: str,
    file: str,
    version: str,
    archive: str,
    checksum: str,
) -> tuple[str, str, int, int, str, float, str, int, float, int, str]:
    r"""Return values of a media entry in dependencies.

    The dependency table expects the following columns:

    * file
    * archive
    * bit depth
    * channels
    * checksum
    * duration
    * format
    * removed
    * sampling rate
    * dependency type
    * version

    Args:
        root: root of database
        file: relative media file path
        version: database version
        archive: archive the media file is stored in
        checksum: checksum of the media file

    Returns:
        row to be added to the dependency table as tuple

    """
    dependency_type = define.DEPENDENCY_TYPE["media"]
    format = audeer.file_extension(file).lower()
    removed = 0

    # Inspect media file to get audio/video metadata
    try:
        path = os.path.join(root, file)
        bit_depth = audiofile.bit_depth(path)
        if bit_depth is None:  # pragma: nocover (non SND files)
            bit_depth = 0
        channels = audiofile.channels(path)
        duration = audiofile.duration(path, sloppy=True)
        sampling_rate = audiofile.sampling_rate(path)
    except FileNotFoundError:  # pragma: nocover
        # If sox or mediafile are not installed
        # we get a FileNotFoundError error
        raise RuntimeError(
            f"sox and mediainfo have to be installed "
            f"to publish '{format}' media files."
        )
    except RuntimeError:
        # Skip audio/video metadata for media files,
        # that don't support them
        # (e.g. text files)
        bit_depth = 0
        channels = 0
        duration = 0.0
        sampling_rate = 0

    return (
        file,
        archive,
        bit_depth,
        channels,
        checksum,
        duration,
        format,
        removed,
        sampling_rate,
        dependency_type,
        version,
    )


def _put_attachments(
    attachments: list[str],
    db_root: str,
    db: audformat.Database,
    version: str,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    def job(attachment_id: str):
        archive_file = backend_interface.join(
            "/", db.name, "attachment", attachment_id + ".zip"
        )
        files = db.attachments[attachment_id].files
        backend_interface.put_archive(db_root, archive_file, version, files=files)

    audeer.run_tasks(
        job,
        params=[([attachment_id], {}) for attachment_id in attachments],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Put attachments",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def _put_media(
    media_archives: set[str],
    db_root: str,
    db_name: str,
    version: str,
    previous_version: str | None,
    deps: Dependencies,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    r"""Upload archives with new, altered or removed media files."""
    if media_archives:
        # create a mapping from archives to media files
        map_archive_to_files = collections.defaultdict(list)
        for file in deps.media:
            if not deps.removed(file):
                map_archive_to_files[deps.archive(file)].append(file)

        def job(archive):
            if archive in map_archive_to_files:
                files = map_archive_to_files[archive]
                for file in files:
                    update_media.append(file)

                archive_file = backend_interface.join(
                    "/", db_name, "media", archive + ".zip"
                )

                if previous_version is not None:
                    # if only some files of the archive were altered
                    # it may happen that the others do not exist
                    # in the root folder,
                    # so we have to download the archive
                    # and copy the missing files first
                    missing_files = []
                    for file in files:
                        path = os.path.join(db_root, file)
                        if not os.path.exists(path):
                            missing_files.append(file)
                    if missing_files:
                        with tempfile.TemporaryDirectory() as tmp_root:
                            backend_interface.get_archive(
                                archive_file,
                                tmp_root,
                                deps.version(missing_files[0]),
                            )
                            for missing_file in missing_files:
                                src_path = os.path.join(tmp_root, missing_file)
                                dst_path = os.path.join(db_root, missing_file)
                                audeer.mkdir(os.path.dirname(dst_path))
                                shutil.copy(
                                    src_path,
                                    dst_path,
                                )

                backend_interface.put_archive(
                    db_root,
                    archive_file,
                    version,
                    files=files,
                )

        update_media = []
        audeer.run_tasks(
            job,
            params=[([archive], {}) for archive in media_archives],
            num_workers=num_workers,
            progress_bar=verbose,
            task_description="Put media",
            maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
        )
        deps._update_media_version(update_media, version)


def _put_tables(
    tables: list[str],
    db_root: str,
    db_name: str,
    version: str,
    backend_interface: type[audbackend.interface.Base],
    num_workers: int | None,
    verbose: bool,
):
    def job(table: str):
        if os.path.exists(os.path.join(db_root, f"db.{table}.parquet")):
            file = os.path.join(db_root, f"db.{table}.parquet")
            remote_file = backend_interface.join(
                "/", db_name, "meta", f"{table}.parquet"
            )
            backend_interface.put_file(file, remote_file, version)
        else:
            file = f"db.{table}.csv"
            archive_file = backend_interface.join("/", db_name, "meta", f"{table}.zip")
            backend_interface.put_archive(db_root, archive_file, version, files=file)

    audeer.run_tasks(
        job,
        params=[([table], {}) for table in tables],
        num_workers=num_workers,
        progress_bar=verbose,
        task_description="Put tables",
        maximum_refresh_time=define.MAXIMUM_REFRESH_TIME,
    )


def publish(
    db_root: str,
    version: str,
    repository: Repository,
    *,
    archives: Mapping[str, str] = None,
    previous_version: str | None = "latest",
    cache_root: str = None,
    num_workers: int | None = 1,
    verbose: bool = True,
) -> Dependencies:
    r"""Publish database.

    Publishes a database conform to audformat_,
    stored in the ``db_root`` folder.

    A database can have dependencies
    to media files and tables of an older version.
    E.g. you might alter an existing table
    by adding labels for new media files to it
    and publish it as a new version.
    :func:`audb.publish` will then upload
    new and altered files and update
    the dependencies accordingly.

    To update a database,
    you first have to load the version
    that the new version should depend on
    with :func:`audb.load_to` to ``db_root``.
    Media files that are not altered can be omitted,
    so it is recommended to set
    ``only_metadata=True`` in :func:`audb.load_to`.
    Afterwards you make your changes to that folder
    and run :func:`audb.publish`.
    To remove media files from a database,
    make sure they are no
    longer referenced in the tables.

    Setting ``previous_version=None`` allows you
    to start from scratch and upload all files
    even if an older versions exist.
    In this case you don't call :func:`audb.load_to`
    before running :func:`audb.publish`.

    Handling of audio formats
    is based on the file extension
    in :mod:`audb`.
    This means the file extension must be lowercase
    and should match the audio format of the file,
    e.g. ``.wav``.

    When canceling :func:`audb.publish`
    during publication
    you can restart it afterwards.
    It will continue from the current state,
    but you might need overwrite permissions
    in addition to write permissions
    on the backend.

    :mod:`audb` uses md5 hashes of the database files
    to check if they have changed.
    Be aware that for certain file formats,
    like parquet,
    md5 hashes might differ
    for files with identical content.
    Reasons include the library that wrote the file,
    involved compression codes,
    or additional metadata written by the library.
    For files stored in parquet format,
    :func:`audb.publish` will first look for a hash
    stored in its metadata
    under the ``b"hash"`` key.
    For parquet tables,
    this deterministic hash
    is automatically added by :mod:`audformat`.

    Tables stored only as pickle files,
    are converted to parquet files
    before publication.
    If a table is stored as a parquet and csv file,
    the csv file is ignored,
    and the parquet file is published.

    .. _audformat: https://audeering.github.io/audformat/data-introduction.html

    Args:
        db_root: root directory of database
        version: version string
        repository: name of repository
        archives: dictionary mapping files to archive names.
            Can be used to bundle files into archives,
            which will speed up communication with the server
            if the database contains many small files.
            Archive name must not include an extension
        previous_version: specifies the version
            this publication should be based on.
            If ``'latest'``
            it will use automatically the latest published version
            or ``None``
            if no version was published.
            If ``None`` it assumes you start from scratch
        cache_root: cache folder where databases are stored.
            If not set :meth:`audb.default_cache_root` is used.
            Only used to read the dependencies of the previous version
        num_workers: number of parallel jobs or 1 for sequential
            processing. If ``None`` will be set to the number of
            processors on the machine multiplied by 5
        verbose: show debug messages

    Returns:
        dependency object

    Raises:
        RuntimeError: if version already exists
        RuntimeError: if database tables reference non-existing files
        RuntimeError: if database attachment path does not exist,
            is a symlink,
            is empty,
            or contains an empty sub-folder
        RuntimeError: if database in ``db_root`` depends on other version
            as indicated by ``previous_version``
        RuntimeError: if database is not portable,
            see :meth:`audformat.Database.is_portable`
        RuntimeError: if non-standard formats like MP3 and MP4 are published,
            but sox and/or mediafile is not installed
        RuntimeError: if the type of a database file changes,
            e.g. from media to attachment
        RuntimeError: if a new media file
            has an uppercase letter in its file extension
        RuntimeError: if database contains tables,
            misc tables, or attachments
            that are stored under an ID
            using a char not in ``'[A-Za-z0-9._-]'``
        ValueError: if ``version`` or ``previous_version``
            cannot be parsed by :class:`audeer.StrictVersion`
        ValueError: if ``previous_version`` >= ``version``
        ValueError: if ``repository`` has artifactory as backend in Python>=3.12
        ValueError: if ``repository`` has a non-supported backend

    """
    # Enforce error if version cannot be converted to audeer.StrictVersion
    audeer.StrictVersion(version)
    if (
        previous_version is not None
        and previous_version != "latest"
        and (audeer.StrictVersion(version) <= audeer.StrictVersion(previous_version))
    ):
        raise ValueError(
            "'previous_version' needs to be smaller than 'version', "
            f"but yours is {previous_version} >= {version}."
        )

    db = audformat.Database.load(
        db_root,
        load_data=False,
        num_workers=num_workers,
        verbose=verbose,
    )

    backend_interface = repository.create_backend_interface()

    with backend_interface.backend:
        remote_header = backend_interface.join("/", db.name, define.HEADER_FILE)
        versions = backend_interface.versions(
            remote_header,
            suppress_backend_errors=True,
        )

    if version in versions:
        raise RuntimeError(
            "A version " f"'{version}' " "already exists for database " f"'{db.name}'."
        )
    if previous_version == "latest":
        if len(versions) > 0:
            previous_version = versions[-1]
        else:
            previous_version = None

    # load database and dependencies
    deps = Dependencies()
    for deps_file in [define.DEPENDENCY_FILE, define.LEGACY_DEPENDENCY_FILE]:
        deps_path = os.path.join(db_root, deps_file)
        if os.path.exists(deps_path):
            deps.load(deps_path)
            break

    # check if database folder depends on the right version

    # dependencies shouldn't be there
    if previous_version is None and len(deps) > 0:
        raise RuntimeError(
            f"You did not set a dependency to a previous version, "
            f"but you have a '{deps_file}' file present "
            f"in {db_root}."
        )

    # dependencies missing
    if previous_version is not None and len(deps) == 0:
        raise RuntimeError(
            f"You want to depend on '{previous_version}' "
            f"of {db.name}, "
            f"but you don't have a '{define.DEPENDENCY_FILE}' file present "
            f"in {db_root}. "
            f"Did you forgot to call "
            f"'audb.load_to({db_root}, {db.name}, "
            f"version={previous_version}?"
        )

    # dependencies do not match version
    if previous_version is not None and len(deps) > 0:
        previous_deps = dependencies(
            db.name,
            version=previous_version,
            cache_root=cache_root,
            verbose=verbose,
        )
        if not deps().equals(previous_deps()):
            raise RuntimeError(
                f"You want to depend on '{previous_version}' "
                f"of {db.name}, "
                f"but the dependency file '{deps_file}' "
                f"in {db_root} "
                f"does not match the dependency file "
                f"for the requested version in the repository. "
                f"Did you forgot to call "
                f"'audb.load_to({db_root}, {db.name}, "
                f"version='{previous_version}') "
                f"or modified the file manually?"
            )

    # load database with table data
    db = audformat.Database.load(
        db_root,
        load_data=True,
        num_workers=num_workers,
        verbose=verbose,
    )

    # ensure table and attachment IDs
    # contain only chars allowed by the backend
    # as the IDs are included in the filenames of the archives
    # uploaded to the backend
    allowed_chars = audbackend.core.utils.BACKEND_ALLOWED_CHARS
    allowed_chars = allowed_chars.replace("/", "")
    allowed_chars_compiled = re.compile(allowed_chars)
    for table_id in list(db):
        if allowed_chars_compiled.fullmatch(table_id) is None:
            raise RuntimeError(
                "Table IDs must only contain chars from "
                f"{allowed_chars[:-1]}, "
                f"which is not the case for table '{table_id}'."
            )
    for attachment_id in list(db.attachments):
        if allowed_chars_compiled.fullmatch(attachment_id) is None:
            raise RuntimeError(
                "Attachment IDs must only contain chars from "
                f"{allowed_chars[:-1]}, "
                f"which is not the case for attachment '{attachment_id}'."
            )

    # check all tables are conform with audformat
    if not db.is_portable:
        raise RuntimeError(
            "Some files in the tables have absolute paths "
            "or use '\\', '.', '..' in its path. "
            "Please replace those paths by relative paths, "
            "use folder names instead of dots, "
            "and avoid Windows path notation."
        )
    _check_for_duplicates(db, num_workers, verbose)

    # check all media referenced in a table exist
    # on disk or are already part of the database
    db_root_files = _get_root_files(db_root)
    _check_for_missing_media(db, db_root, db_root_files, deps)

    # Make sure all tables are stored in CSV or PARQUET format.
    # If only a PKL is found,
    # the table is stored as PARQUET instead
    for table_id in list(db):
        table = db[table_id]
        table_path = os.path.join(db_root, f"db.{table_id}")
        if not os.path.exists(f"{table_path}.csv") and not os.path.exists(
            f"{table_path}.parquet"
        ):
            table.save(table_path, storage_format="parquet")

    # check archives
    archives = archives or {}

    with backend_interface.backend:
        # publish attachments
        attachments = _find_attachments(db, db_root, version, deps, verbose)
        _put_attachments(
            attachments, db_root, db, version, backend_interface, num_workers, verbose
        )

        # publish tables
        tables = _find_tables(db, db_root, version, deps, verbose)
        _put_tables(
            tables, db_root, db.name, version, backend_interface, num_workers, verbose
        )

        # publish media
        media_archives = _find_media(
            db, db_root, db_root_files, version, deps, archives, num_workers, verbose
        )
        _put_media(
            media_archives,
            db_root,
            db.name,
            version,
            previous_version,
            deps,
            backend_interface,
            num_workers,
            verbose,
        )

        # publish dependencies and header
        upload_dependencies(backend_interface, deps, db_root, db.name, version)
        try:
            local_header = os.path.join(db_root, define.HEADER_FILE)
            remote_header = backend_interface.join("/", db.name, define.HEADER_FILE)
            backend_interface.put_file(local_header, remote_header, version)
        except Exception:  # pragma: no cover
            # after the header is published
            # the new version becomes visible,
            # so if something goes wrong here
            # we better clean up
            if backend_interface.exists(remote_header, version):
                backend_interface.remove_file(remote_header, version)

    return deps



import os

import numpy as np
import pytest

import audeer
import audformat.testing
import audiofile

import audb


DB_NAME = "test_convert"


@pytest.fixture(
    scope="module",
    autouse=True,
)
def db_root(tmpdir_factory, persistent_repository):
    r"""Publish single database.

    Returns:
        path to original database root

    """
    version = "1.0.0"
    db_root = tmpdir_factory.mktemp(version)

    # define audio files and metadata

    db_files = {
        "audio/file1.wav": {
            "bit_depth": 16,
            "channels": 1,
            "format": "wav",
            "sampling_rate": 8000,
        },
        "audio/file2.wav": {
            "bit_depth": 24,
            "channels": 2,
            "format": "wav",
            "sampling_rate": 16000,
        },
        "audio/file3.flac": {
            "bit_depth": 8,
            "channels": 3,
            "format": "flac",
            "sampling_rate": 44100,
        },
    }

    # create db

    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME

    db["files"] = audformat.Table(audformat.filewise_index(list(db_files)))
    db["files"]["original"] = audformat.Column()
    db["files"]["original"].set(list(db_files))
    for file in db_files:
        signal = np.zeros(
            (
                db_files[file]["channels"],
                db_files[file]["sampling_rate"],
            ),
            dtype=np.float32,
        )
        path = os.path.join(db_root, file)
        audeer.mkdir(os.path.dirname(path))
        audiofile.write(
            path,
            signal,
            db_files[file]["sampling_rate"],
            bit_depth=db_files[file]["bit_depth"],
        )
    db["segments"] = audformat.Table(
        audformat.segmented_index(
            [list(db_files)[0]] * 3,
            starts=["0s", "1s", "2s"],
            ends=["1s", "2s", "3s"],
        )
    )
    db.save(db_root)

    # publish db

    audb.publish(
        db_root,
        version,
        persistent_repository,
        verbose=False,
    )

    return db_root


@pytest.mark.parametrize(
    "bit_depth",
    [
        None,
        16,
    ],
)
def test_bit_depth(db_root, bit_depth):
    db = audb.load(
        DB_NAME,
        bit_depth=bit_depth,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    original_files = db["files"]["original"].get()

    df = audb.cached()
    assert df["bit_depth"].values[0] == bit_depth

    for converted_file, original_file in zip(db.files, original_files):
        converted_file = os.path.join(db.meta["audb"]["root"], converted_file)
        original_file = os.path.join(db_root, original_file)

        if bit_depth is None:
            assert audiofile.bit_depth(converted_file) == audiofile.bit_depth(
                original_file
            )
        else:
            assert audiofile.bit_depth(converted_file) == bit_depth


@pytest.mark.parametrize(
    "channels",
    [
        None,
        1,
        [0, -1],
        range(5),
    ],
)
def test_channels(db_root, channels):
    db = audb.load(
        DB_NAME,
        channels=channels,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    original_files = db["files"]["original"].get()

    df = audb.cached()

    for converted_file, original_file in zip(db.files, original_files):
        converted_file = os.path.join(db.meta["audb"]["root"], converted_file)
        original_file = os.path.join(db_root, original_file)

        if channels is None:
            assert audiofile.channels(converted_file) == audiofile.channels(
                original_file
            )
            assert df["channels"].values[0] == channels
        elif isinstance(channels, int):
            assert audiofile.channels(converted_file) == 1
            assert df["channels"].values[0] == [1]
        else:
            assert audiofile.channels(converted_file) == len(channels)


@pytest.mark.parametrize(
    "format",
    [
        None,
        audb.core.define.Format.WAV,
        audb.core.define.Format.FLAC,
    ],
)
def test_format(db_root, format):
    db = audb.load(
        DB_NAME,
        format=format,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    original_files = db["files"]["original"].get()

    df = audb.cached()
    assert df["format"].values[0] == format

    for converted_file, original_file in zip(db.files, original_files):
        converted_file = os.path.join(db.meta["audb"]["root"], converted_file)
        original_file = os.path.join(db_root, original_file)

        if format is None:
            assert converted_file[-4:] == original_file[-4:]
        else:
            assert converted_file.endswith(format)


@pytest.mark.parametrize(
    "mixdown",
    [False, True],
)
def test_mixdown(db_root, mixdown):
    db = audb.load(
        DB_NAME,
        mixdown=mixdown,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    original_files = db["files"]["original"].get()

    df = audb.cached()
    assert df["mixdown"].values[0] == mixdown

    for converted_file, original_file in zip(db.files, original_files):
        converted_file = os.path.join(db.meta["audb"]["root"], converted_file)
        original_file = os.path.join(db_root, original_file)

        if mixdown:
            assert audiofile.channels(converted_file) == 1
        else:
            assert audiofile.channels(converted_file) == audiofile.channels(
                original_file
            )


@pytest.mark.parametrize(
    "sampling_rate",
    [None, 16000, 24000],
)
def test_sampling_rate(db_root, sampling_rate):
    db = audb.load(
        DB_NAME,
        sampling_rate=sampling_rate,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    original_files = db["files"]["original"].get()

    df = audb.cached()
    assert df["sampling_rate"].values[0] == sampling_rate

    for converted_file, original_file in zip(db.files, original_files):
        converted_file = os.path.join(db.meta["audb"]["root"], converted_file)
        original_file = os.path.join(db_root, original_file)

        if sampling_rate is None:
            assert audiofile.sampling_rate(converted_file) == audiofile.sampling_rate(
                original_file
            )
        else:
            assert audiofile.sampling_rate(converted_file) == sampling_rate


def test_mixed_cache(cache, shared_cache):
    # Avoid failing searching for other versions
    # if databases a stored across private and shared cache
    # and the private one is empty, see
    # https://github.com/audeering/audb/issues/101

    # First load to shared cache
    audb.load(
        DB_NAME,
        sampling_rate=8000,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
        only_metadata=True,
        tables="files",
        cache_root=shared_cache,
    )
    # Now try to load same version to private cache
    # to force audb.cached() to return empty dataframe
    audeer.rmdir(cache)
    audb.load(
        DB_NAME,
        sampling_rate=8000,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
        only_metadata=True,
        tables="segments",
    )



import argparse

import audb


def main(pandas_version):
    """Load dependency from CSV and store as PKL file.

    Args:
        pandas_version: version of installed ``pandas`` package

    """
    # Download emodb dependency table
    # from version 1.4.1,
    # which is still stored as CSV file
    repository = audb.Repository(
        "data-public",
        "https://audeering.jfrog.io/artifactory",
        "artifactory",
    )
    backend_interface = repository.create_backend_interface()
    remote_file = backend_interface.join("/", "emodb", "db.zip")
    with backend_interface.backend:
        backend_interface.get_archive(remote_file, ".", "1.4.1", verbose=False)

    deps = audb.Dependencies()
    deps.load("db.csv")
    outfile = f"emodb-pandas-{pandas_version}.pkl"
    deps.save(outfile)


if __name__ == "__main__":
    # Call the program with:
    #
    # $ python store_dependency_table.py 2.2.2
    #
    # where 2.2.2 refers to the installed pandas version.
    parser = argparse.ArgumentParser()
    parser.add_argument("pandas_version")
    args = parser.parse_args()
    main(args.pandas_version)



import os
import random
import shutil

import numpy as np
import pandas as pd
import pytest

import audeer
import audformat.testing
import audiofile

import audb


DB_NAME = "test_load"


@pytest.fixture(
    scope="function",
    autouse=True,
)
def assert_database_tmp_folder_is_deleted():
    """Fixture to test that the ~ tmp folder gets deleted.

    audb.load() first loads files to a folder
    named after the database
    and appended by ``'~'``.
    This folder should be deleted in the end.

    """
    yield

    dirs = audeer.list_dir_names(audb.default_cache_root(), recursive=True)
    assert len([d for d in dirs if d.endswith("~")]) == 0


@pytest.fixture(
    scope="module",
    autouse=True,
)
def dbs(tmpdir_factory, persistent_repository, storage_format):
    r"""Publish different versions of the same database.

    Returns:
        dictionary containing root folder for each version

    """
    # Fix seed for audformat.testing
    random.seed(1)

    # Collect single database paths
    # and return them in the end
    paths = {}

    # create db

    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db.schemes["scheme"] = audformat.Scheme(labels=["positive", "neutral", "negative"])
    audformat.testing.add_table(
        db,
        "emotion",
        audformat.define.IndexType.SEGMENTED,
        num_files=5,
        columns={"emotion": ("scheme", None)},
    )
    audformat.testing.add_misc_table(
        db,
        "misc-in-scheme",
        pd.Index([0, 1, 2], dtype="Int64", name="idx"),
        columns={"emotion": ("scheme", None)},
    )
    db.schemes["speaker"] = audformat.Scheme(labels=["adam", "eve"])
    db.schemes["misc"] = audformat.Scheme(
        "int",
        labels="misc-in-scheme",
    )
    db["files"] = audformat.Table(db.files)
    db["files"]["speaker"] = audformat.Column(scheme_id="speaker")
    db["files"]["speaker"].set(
        ["adam", "adam", "eve", "eve"],
        index=audformat.filewise_index(db.files[:4]),
    )
    db["files"]["misc"] = audformat.Column(scheme_id="misc")
    db["files"]["misc"].set(
        [0, 1, 1, 2],
        index=audformat.filewise_index(db.files[:4]),
    )
    db.attachments["file"] = audformat.Attachment("extra/file.txt")
    db.attachments["folder"] = audformat.Attachment("extra/folder")

    # publish 1.0.0

    version = "1.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)

    audeer.mkdir(db_root, "extra/folder/sub-folder")
    audeer.touch(db_root, "extra/file.txt")
    audeer.touch(db_root, "extra/folder/file1.txt")
    audeer.touch(db_root, "extra/folder/file2.txt")
    audeer.touch(db_root, "extra/folder/sub-folder/file3.txt")
    db.save(db_root, storage_format=storage_format)
    audformat.testing.create_audio_files(db)
    archives = db["files"]["speaker"].get().dropna().to_dict()
    audb.publish(
        db_root,
        version,
        persistent_repository,
        archives=archives,
        verbose=False,
    )

    # publish 1.1.0, add table, remove attachment file

    previous_db_root = db_root
    version = "1.1.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)

    audformat.testing.add_table(
        db,
        "train",
        audformat.define.IndexType.SEGMENTED,
        columns={"label": ("scheme", None)},
    )
    shutil.copytree(
        audeer.path(previous_db_root, "extra"),
        audeer.path(db_root, "extra"),
    )
    os.remove(audeer.path(db_root, "extra/folder/file2.txt"))

    db.save(db_root, storage_format=storage_format)
    audformat.testing.create_audio_files(db)
    shutil.copy(
        audeer.path(previous_db_root, audb.core.define.DEPENDENCY_FILE),
        audeer.path(db_root, audb.core.define.DEPENDENCY_FILE),
    )
    audb.publish(
        db_root,
        version,
        persistent_repository,
        verbose=False,
    )

    # publish 1.1.1, change label

    previous_db_root = db_root
    version = "1.1.1"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)

    row = db["train"].index[0]
    db["train"].df.at[row, "label"] = None
    shutil.copytree(
        audeer.path(previous_db_root, "extra"),
        audeer.path(db_root, "extra"),
    )

    db.save(db_root, storage_format=storage_format)
    audformat.testing.create_audio_files(db)
    shutil.copy(
        audeer.path(previous_db_root, audb.core.define.DEPENDENCY_FILE),
        audeer.path(db_root, audb.core.define.DEPENDENCY_FILE),
    )
    audb.publish(
        db_root,
        version,
        persistent_repository,
        verbose=False,
    )

    # publish 2.0.0, alter and remove media, remove attachment

    previous_db_root = db_root
    version = "2.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)

    shutil.copytree(
        audeer.path(previous_db_root, "extra"),
        audeer.path(db_root, "extra"),
    )
    del db.attachments["file"]
    os.remove(audeer.path(db_root, "extra/file.txt"))

    db.save(db_root, storage_format=storage_format)
    audformat.testing.create_audio_files(db)
    file = os.path.join(db_root, db.files[0])
    y, sr = audiofile.read(file)
    y[0] = 1
    audiofile.write(file, y, sr)
    file = db.files[-1]
    db.pick_files(lambda x: x != file)
    os.remove(audeer.path(db_root, file))
    db.save(db_root, storage_format=storage_format)

    shutil.copy(
        os.path.join(previous_db_root, audb.core.define.DEPENDENCY_FILE),
        os.path.join(db_root, audb.core.define.DEPENDENCY_FILE),
    )
    audb.publish(
        db_root,
        version,
        persistent_repository,
        verbose=False,
    )

    # publish 3.0.0, remove table, alter attachment file

    previous_db_root = db_root
    version = "3.0.0"
    db_root = tmpdir_factory.mktemp(version)
    paths[version] = str(db_root)

    shutil.copytree(
        audeer.path(previous_db_root, "extra"),
        audeer.path(db_root, "extra"),
    )
    with open(audeer.path(db_root, "extra/folder/file1.txt"), "a") as fp:
        fp.write("text")

    db.drop_tables("train")
    db.save(db_root, storage_format=storage_format)
    audformat.testing.create_audio_files(db)
    shutil.copy(
        os.path.join(previous_db_root, audb.core.define.DEPENDENCY_FILE),
        os.path.join(db_root, audb.core.define.DEPENDENCY_FILE),
    )
    audb.publish(
        db_root,
        version,
        persistent_repository,
        verbose=False,
    )

    return paths


def test_database_cache_folder(cache):
    cache_root = os.path.join(cache, "cache")
    version = "1.0.0"
    db_root = audb.core.load.database_cache_root(
        DB_NAME,
        version,
        cache_root,
    )
    expected_db_root = os.path.join(
        cache_root,
        DB_NAME,
        version,
    )
    assert db_root == expected_db_root


def test_load_wrong_argument():
    with pytest.raises(TypeError):
        audb.load(DB_NAME, typo="1.0.0")


@pytest.mark.parametrize("only_metadata", [True, False])
@pytest.mark.parametrize(
    "format",
    [
        None,
        "wav",
        "flac",
    ],
)
@pytest.mark.parametrize(
    "version",
    [
        None,  # 3.0.0
        "1.0.0",
        "1.1.0",
        "1.1.1",
        "2.0.0",
        pytest.param(
            "4.0.0",
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
    ],
)
def test_load(dbs, format, version, only_metadata):
    # When loading the first time (only_metadata=True)
    # the database should not exists in cache
    if only_metadata:
        assert not audb.exists(
            DB_NAME,
            version=version,
            format=format,
        )

    # === Load from REPOSITORY with full_path=False ===

    db = audb.load(
        DB_NAME,
        version=version,
        format=format,
        only_metadata=only_metadata,
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    db_root = db.meta["audb"]["root"]

    # Load original database from folder (expected database)
    resolved_version = version or audb.latest_version(DB_NAME)
    db_original = audformat.Database.load(dbs[resolved_version])
    if format is not None:
        db_original.map_files(lambda x: audeer.replace_file_extension(x, format))

    # Assert database exists in cache
    assert audb.exists(DB_NAME, version=version, format=format)
    df = audb.cached()
    assert df.loc[db_root, "version"] == resolved_version

    # Assert files duration are stored as hidden attribute
    if not only_metadata:
        files_duration = {
            os.path.join(db_root, os.path.normpath(file)): pd.to_timedelta(
                audiofile.duration(os.path.join(db_root, file)), unit="s"
            )
            for file in db.files
        }
        assert db._files_duration == files_duration

    # Assert media files are identical and (not) exist
    pd.testing.assert_index_equal(db.files, db_original.files)
    for file in db.files:
        if only_metadata:
            assert not os.path.exists(os.path.join(db_root, file))
        else:
            assert os.path.exists(os.path.join(db_root, file))

    # Assert tables are identical and table files exist
    for table in db:
        assert os.path.exists(os.path.join(db_root, f"db.{table}.csv"))
        pd.testing.assert_frame_equal(
            db_original[table].df,
            db[table].df,
        )

    # Assert attachments are identical and files do (not) exist
    assert db.attachments == db_original.attachments
    for attachment in db.attachments:
        path = audeer.path(db.root, db.attachments[attachment].path)
        if only_metadata:
            assert not os.path.exists(path)
        else:
            assert os.path.exists(path)
            for attachment_file in db.attachments[attachment].files:
                assert os.path.exists(audeer.path(db.root, attachment_file))

    # Assert all files are listed in dependency table
    deps = audb.dependencies(DB_NAME, version=version)
    assert len(deps) == (
        len(db.files) + len(db.tables) + len(db.misc_tables) + len(db.attachments)
    )

    # === Load from CACHE with full_path=True ===

    db = audb.load(
        DB_NAME,
        version=version,
        full_path=True,
        format=format,
        only_metadata=only_metadata,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )

    # Assert files duration are stored as hidden attribute
    if not only_metadata:
        files_duration = {
            os.path.normpath(file): pd.to_timedelta(audiofile.duration(file), unit="s")
            for file in db.files
        }
        assert db._files_duration == files_duration

    # Assert media files do (not) exist
    for file in db.files:
        if only_metadata:
            assert not os.path.exists(file)
        else:
            assert os.path.exists(file)

    # Assert table files exist
    for table in db:
        assert os.path.exists(os.path.join(db_root, f"db.{table}.csv"))

    # Assert attachments files do (not) exist
    for attachment in db.attachments:
        path = audeer.path(db.root, db.attachments[attachment].path)
        if only_metadata:
            assert not os.path.exists(path)
        else:
            assert os.path.exists(path)
            for attachment_file in db.attachments[attachment].files:
                assert os.path.exists(audeer.path(db.root, attachment_file))


def test_load_from_cache(dbs):
    # Load a database with flavor to cache
    # and reload afterwards from cache
    format = "flac"
    version = "1.0.0"
    db = audb.load(
        DB_NAME,
        version="1.0.0",
        format="flac",
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    db_root = db.meta["audb"]["root"]

    # Load original database from folder (expected database)
    db_original = audformat.Database.load(dbs[version])
    db_original.map_files(lambda x: audeer.replace_file_extension(x, format))

    # Assert database exists in cache
    assert audb.exists(DB_NAME, version=version, format=format)
    df = audb.cached()
    assert df.loc[db_root, "version"] == version

    # Assert media files are identical and exist
    pd.testing.assert_index_equal(db.files, db_original.files)
    for file in db.files:
        assert os.path.exists(os.path.join(db_root, file))

    # Ensure no media files in flavor cache are marked as missing files,
    # when flavor format is different from original format
    # (https://github.com/audeering/audb/issues/324)
    original_files = audformat.utils.replace_file_extension(db.files, "wav")
    assert (
        audb.core.load._missing_files(
            original_files,
            "media",
            db_root,
            audb.Flavor(format="flac"),
            False,
        )
        == []
    )

    version = "2.0.0"
    db = audb.load(
        DB_NAME,
        version="2.0.0",
        format="flac",
        full_path=False,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    db_root = db.meta["audb"]["root"]

    # Load original database from folder (expected database)
    db_original = audformat.Database.load(dbs[version])
    db_original.map_files(lambda x: audeer.replace_file_extension(x, format))

    # Assert database exists in cache
    assert audb.exists(DB_NAME, version=version, format=format)
    df = audb.cached()
    assert df.loc[db_root, "version"] == version

    # Assert media files are identical and exist
    pd.testing.assert_index_equal(db.files, db_original.files)
    for file in db.files:
        assert os.path.exists(os.path.join(db_root, file))


@pytest.mark.parametrize(
    "version, attachment_id",
    [
        (
            "1.0.0",
            "file",
        ),
        (
            "1.0.0",
            "folder",
        ),
        (
            None,
            "folder",
        ),
    ],
)
def test_load_attachment(cache, version, attachment_id):
    db = audb.load(
        DB_NAME,
        version=version,
        verbose=False,
    )

    paths = audb.load_attachment(
        DB_NAME,
        attachment_id,
        version=version,
        verbose=False,
    )

    if version is None:
        version = audb.latest_version(DB_NAME)

    expected_paths = [
        os.path.join(
            cache,
            DB_NAME,
            version,
            os.path.normpath(file),
        )
        for file in db.attachments[attachment_id].files
    ]
    assert paths == expected_paths

    # Clear cache to force loading from other cache
    cache_root = audb.core.load.database_cache_root(
        DB_NAME,
        version,
        cache,
        audb.Flavor(),
    )
    shutil.rmtree(cache_root)
    paths2 = audb.load_attachment(
        DB_NAME,
        attachment_id,
        version=version,
        verbose=False,
    )
    assert paths2 == expected_paths


@pytest.mark.parametrize(
    "version, attachment_id, error, error_msg",
    [
        ("1.0.0", "", ValueError, "Could not find the attachment ''"),
        (
            "1.0.0",
            "non-existent",
            ValueError,
            "Could not find the attachment 'non-existent'",
        ),
    ],
)
def test_load_attachment_errors(version, attachment_id, error, error_msg):
    with pytest.raises(error, match=error_msg):
        audb.load_attachment(
            DB_NAME,
            attachment_id,
            version=version,
            verbose=False,
        )


@pytest.mark.parametrize(
    "version, media, format",
    [
        (
            "1.0.0",
            [],
            None,
        ),
        (
            "1.0.0",
            "audio/001.wav",
            "wav",
        ),
        pytest.param(
            "1.0.0",
            ["audio/001.flac", "audio/002.flac"],
            "flac",
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        (
            "1.0.0",
            ["audio/001.wav", "audio/002.wav"],
            "flac",
        ),
        (
            None,
            ["audio/001.wav"],
            None,
        ),
    ],
)
def test_load_media(cache, version, media, format):
    paths = audb.load_media(
        DB_NAME,
        media,
        version=version,
        format=format,
        verbose=False,
    )
    expected_paths = [os.path.join(cache, p) for p in paths]
    if format is not None:
        expected_paths = [
            audeer.replace_file_extension(p, format) for p in expected_paths
        ]
    assert paths == expected_paths

    # Clear cache to force loading from other cache
    if version is None:
        version = audb.latest_version(DB_NAME)
    cache_root = audb.core.load.database_cache_root(
        DB_NAME,
        version,
        cache,
        audb.Flavor(format=format),
    )
    shutil.rmtree(cache_root)
    paths2 = audb.load_media(
        DB_NAME,
        media,
        version=version,
        format=format,
        verbose=False,
    )
    assert paths2 == paths


@pytest.mark.parametrize("pickle_tables", [True, False])
@pytest.mark.parametrize("name, version, table", [(DB_NAME, "1.0.0", "emotion")])
class TestLoadPickle:
    r"""Test storing tables as pickle files in cache.

    When tables are first downloaded from a backend,
    they are stored in their original format in the cache,
    and dependent on the ``pickle_tables`` argument
    stored as pickle as well.

    """

    def assert_table_exist_in_cache(
        self, db_root, table, storage_format, pickle_tables
    ):
        """Assert table exists in original format and maybe as pickle.

        Args:
            db_root: database root folder
            table: table ID
            storage_format: storage format of table, ``"csv"`` or ``"parquet"``
            pickle_tables: if ``True``,
                table is asserted to exist as pickle file as well,
                otherwise to not exist

        """
        original_table = audeer.path(db_root, f"db.{table}.{storage_format}")
        assert os.path.exists(original_table)

        pickled_table = audeer.path(db_root, f"db.{table}.pkl")
        if pickle_tables:
            assert os.path.exists(pickled_table)
        else:
            assert not os.path.exists(pickled_table)

    def test_load_pickle(self, storage_format, name, version, table, pickle_tables):
        """Test storing tables with audb.load().

        Args:
            storage_format: storage_format fixture
            name: database name
            version: database version
            table: table ID
            pickle_tables: if ``True``,
                tables are stored as pickle files as well in cache

        """
        db = audb.load(
            name,
            version=version,
            tables=table,
            pickle_tables=pickle_tables,
            only_metadata=True,
            verbose=False,
        )
        self.assert_table_exist_in_cache(db.root, table, storage_format, pickle_tables)

    def test_load_table_pickle(
        self, cache, storage_format, name, version, table, pickle_tables
    ):
        r"""Test storing tables with audb.load_table().

        Args:
            cache: cache fixture
            storage_format: storage_format fixture
            name: database name
            version: database version
            table: table ID
            pickle_tables: if ``True``,
                tables are stored as pickle files as well in cache

        """
        audb.load_table(
            name,
            table,
            version=version,
            pickle_tables=pickle_tables,
            verbose=False,
        )
        db_root = audeer.path(cache, name, version)
        self.assert_table_exist_in_cache(db_root, table, storage_format, pickle_tables)


@pytest.mark.parametrize(
    "version, table",
    [
        (
            "1.0.0",
            "emotion",
        ),
        pytest.param(
            "1.0.0",
            "non-existing-table",
            marks=pytest.mark.xfail(raises=ValueError),
        ),
        (
            None,
            "emotion",
        ),
    ],
)
def test_load_table(version, table):
    df = audb.load_table(
        DB_NAME,
        table,
        version=version,
        verbose=False,
    )
    if version is None:
        expected_files = [
            "audio/001.wav",
            "audio/002.wav",
            "audio/003.wav",
            "audio/004.wav",
        ]
    elif version == "1.0.0":
        expected_files = [
            "audio/001.wav",
            "audio/002.wav",
            "audio/003.wav",
            "audio/004.wav",
            "audio/005.wav",
        ]
    files = sorted(list(set(df.index.get_level_values("file"))))
    assert files == expected_files


@pytest.mark.parametrize(
    "version, table, map, expected",
    [
        (
            "1.0.0",
            "files",
            None,
            pd.concat(
                [
                    pd.Series(
                        index=audformat.filewise_index(
                            [f"audio/00{n + 1}.wav" for n in range(5)]
                        ),
                        name="speaker",
                        data=["adam", "adam", "eve", "eve", None],
                        dtype=pd.CategoricalDtype(["adam", "eve"]),
                    ),
                    pd.Series(
                        index=audformat.filewise_index(
                            [f"audio/00{n + 1}.wav" for n in range(5)]
                        ),
                        name="misc",
                        data=[0, 1, 1, 2, np.nan],
                        dtype=pd.CategoricalDtype([0, 1, 2]),
                    ),
                ],
                axis=1,
            ),
        ),
        (
            "1.0.0",
            "files",
            {"misc": "emotion"},
            pd.concat(
                [
                    pd.Series(
                        index=audformat.filewise_index(
                            [f"audio/00{n + 1}.wav" for n in range(5)]
                        ),
                        name="speaker",
                        data=["adam", "adam", "eve", "eve", None],
                        dtype=pd.CategoricalDtype(["adam", "eve"]),
                    ),
                    pd.Series(
                        index=audformat.filewise_index(
                            [f"audio/00{n + 1}.wav" for n in range(5)]
                        ),
                        name="emotion",
                        data=["positive", "positive", "positive", "negative", None],
                        dtype=pd.CategoricalDtype(["positive", "neutral", "negative"]),
                    ),
                ],
                axis=1,
            ),
        ),
    ],
)
def test_load_table_map(version, table, map, expected):
    df = audb.load_table(
        DB_NAME,
        table,
        version=version,
        map=map,
        verbose=False,
    )
    pd.testing.assert_frame_equal(df, expected)


@pytest.mark.parametrize(
    "version",
    [
        None,
        "1.0.0",
        "1.1.0",
        "1.1.1",
        "2.0.0",
        "3.0.0",
        pytest.param(
            "4.0.0",
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
    ],
)
@pytest.mark.parametrize("only_metadata", [True, False])
def test_load_to(tmpdir, dbs, version, only_metadata):
    db_root = audeer.path(tmpdir, "raw")

    db = audb.load_to(
        db_root,
        DB_NAME,
        version=version,
        only_metadata=only_metadata,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert db.root == db_root

    # Load original database from folder (expected database)
    resolved_version = version or audb.latest_version(DB_NAME)
    db_original = audformat.Database.load(dbs[resolved_version])

    # Assert media files are identical and do (not) exist
    pd.testing.assert_index_equal(db.files, db_original.files)
    for file in db.files:
        if only_metadata:
            assert not os.path.exists(os.path.join(db_root, file))
        else:
            assert os.path.exists(os.path.join(db_root, file))

    # Assert tables are identical and exist as files
    for table in db:
        assert os.path.exists(os.path.join(db_root, f"db.{table}.csv"))
        pd.testing.assert_frame_equal(
            db_original[table].df,
            db[table].df,
        )

    # Assert attachments are identical and files exist
    assert db.attachments == db_original.attachments
    for attachment in db.attachments:
        path = audeer.path(db.root, db.attachments[attachment].path)
        if only_metadata:
            assert not os.path.exists(path)
        else:
            assert os.path.exists(path)
            for attachment_file in db.attachments[attachment].files:
                assert os.path.exists(audeer.path(db.root, attachment_file))


@pytest.mark.parametrize("only_metadata", [True, False])
def test_load_to_update(tmpdir, dbs, only_metadata):
    # Use version 1.0.0 as this contains two attachments,
    # one file and one folder
    # which is needed to reach full code coverage
    version = "1.0.0"

    db_root = audeer.path(tmpdir, "raw")

    db = audb.load_to(
        db_root,
        DB_NAME,
        version=version,
        only_metadata=only_metadata,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert db.root == db_root

    # Remove some files
    if not only_metadata:
        media = audeer.path(db_root, db.files[0])
        os.remove(media)
    table = audeer.path(db_root, f"db.{list(db)[0]}.csv")
    os.remove(table)

    # Change some files
    if not only_metadata:
        for attachment_id in list(db.attachments):
            attachment = audeer.path(
                db_root,
                db.attachments[attachment_id].path,
            )
            if os.path.isdir(attachment):
                audeer.touch(attachment, "other-file.txt")
            else:
                with open(attachment, "a") as fp:
                    fp.write("next")

    # Load again to force restoring to original state
    db = audb.load_to(
        db_root,
        DB_NAME,
        version=version,
        only_metadata=only_metadata,
        num_workers=pytest.NUM_WORKERS,
        verbose=False,
    )
    assert db.root == db_root

    # Load original database from folder (expected database)
    resolved_version = version or audb.latest_version(DB_NAME)
    db_original = audformat.Database.load(dbs[resolved_version])

    # Assert media files are identical and do (not) exist
    pd.testing.assert_index_equal(db.files, db_original.files)
    for file in db.files:
        if only_metadata:
            assert not os.path.exists(os.path.join(db_root, file))
        else:
            assert os.path.exists(os.path.join(db_root, file))

    # Assert tables are identical and exist as CSV files
    for table in db:
        assert os.path.exists(os.path.join(db_root, f"db.{table}.csv"))
        pd.testing.assert_frame_equal(
            db_original[table].df,
            db[table].df,
        )

    # Assert attachments are identical and files exist
    assert db.attachments == db_original.attachments
    for attachment in db.attachments:
        path = audeer.path(db.root, db.attachments[attachment].path)
        if only_metadata:
            assert not os.path.exists(path)
        else:
            assert os.path.exists(path)
            for attachment_file in db.attachments[attachment].files:
                assert os.path.exists(audeer.path(db.root, attachment_file))


@pytest.mark.parametrize(
    "name, version, error, error_msg",
    [
        (DB_NAME, "1.0.0", None, None),
        pytest.param(  # database does not exist
            "does-not-exist",
            "1.0.0",
            RuntimeError,
            "Cannot find database 'does-not-exist'.",
        ),
        pytest.param(  # version does not exist
            DB_NAME,
            "999.9.9",
            RuntimeError,
            f"Cannot find version '999.9.9' for database '{DB_NAME}'.",
        ),
    ],
)
def test_repository(persistent_repository, name, version, error, error_msg):
    if error is not None:
        with pytest.raises(error, match=error_msg):
            repository = audb.repository(name, version)
    else:
        repository = audb.repository(name, version)
        assert repository == persistent_repository



import os

import pytest

import audeer
import audformat.testing

import audb


DB_NAME = "test_remove"
DB_FILES = {
    "1.0.0": [
        "audio/bundle1.wav",
        "audio/bundle2.wav",
        "audio/single.wav",
    ],
    "2.0.0": [
        "audio/new.wav",
    ],
}


@pytest.fixture(
    scope="module",
    autouse=True,
)
def dbs(tmpdir_factory, persistent_repository):
    r"""Publish databases.

    This publishes a database with the name ``DB_NAME``
    and the versions 1.0.0 and 2.0.0
    to a module wide repository.

    """
    db_root = tmpdir_factory.mktemp("db")

    # create db

    db = audformat.testing.create_db(minimal=True)
    db.name = DB_NAME
    db["files"] = audformat.Table(audformat.filewise_index(DB_FILES["1.0.0"]))

    # publish 1.0.0

    db.save(db_root)
    audformat.testing.create_audio_files(db)
    archives = {
        db.files[0]: "bundle",
        db.files[1]: "bundle",
    }
    audb.publish(
        db_root,
        "1.0.0",
        persistent_repository,
        archives=archives,
        verbose=False,
    )

    # publish 2.0.0

    db["files"].extend_index(
        audformat.filewise_index(DB_FILES["2.0.0"]),
        inplace=True,
    )
    db.save(db_root)
    audformat.testing.create_audio_files(db)
    audb.publish(
        db_root,
        "2.0.0",
        persistent_repository,
        verbose=False,
    )


@pytest.mark.parametrize(
    "format",
    [
        None,
        "wav",
        "flac",
    ],
)
def test_remove(cache, format):
    for remove in (
        DB_FILES["1.0.0"][0],  # bundle1
        DB_FILES["1.0.0"][1],  # bundle2
        DB_FILES["1.0.0"][2],  # single
        DB_FILES["2.0.0"][0],  # new
    ):
        # remove db cache to ensure we always get a fresh copy
        audeer.rmdir(cache)

        audb.remove_media(DB_NAME, remove)

        for removed_media in [False, True]:
            for version in audb.versions(DB_NAME):
                if remove in DB_FILES[version]:
                    if format is not None:
                        name, _ = os.path.splitext(remove)
                        removed_file = f"{name}.{format}"
                    else:
                        removed_file = remove

                    db = audb.load(
                        DB_NAME,
                        version=version,
                        format=format,
                        removed_media=removed_media,
                        full_path=False,
                        num_workers=pytest.NUM_WORKERS,
                        verbose=False,
                    )

                    if removed_media:
                        assert removed_file in db.files
                    else:
                        assert removed_file not in db.files
                    assert removed_file not in audeer.list_file_names(
                        os.path.join(db.meta["audb"]["root"], "audio"),
                    )

        # Make sure calling it again doesn't raise error
        audb.remove_media(DB_NAME, remove)
